//#define ClientDebug

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Text;
using System.Text.RegularExpressions;
using System.IO;
using System.Security;
using System.Net;
using System.Net.Sockets;
using System.Management.Automation;
using System.Management.Automation.Host;
using System.Management.Automation.Runspaces;
using System.Globalization;
using System.Runtime.InteropServices;
using System.Security.Authentication;
using System.Security.Cryptography.X509Certificates;

namespace PowerShellRemoting
{
    /// <summary>
    /// Wrapper object severs as a proxy between remote sever user interface and local client user interface
    /// of PSHost.
    /// </summary>
    /// <remarks>
    /// PSHostUserInterface object is required to use this class. See http://windowssdk.msdn.microsoft.com/library/default.asp?url=/library/en-us/Monad_Prog/html/f3aaf667-af84-4ea8-a5ad-d454d0d700b8.asp
    /// </remarks>
    /// <example>
    /// static void Main(string[] args)
    /// {
    ///      System.Net.IPEndPoint RemoteEP = new System.Net.IPEndPoint(System.Net.IPAddress.Parse("127.0.0.1"), 8080);
    ///      System.Management.Automation.Host.PSHostUserInterface LocalUI = new LocalPSHostUserInterface();
    ///      ClientPSHost clientHost = new ClientPSHost(RemoteEP, LocalUI);
    ///      clientHost.Start();
    /// }
    /// </example>
    public sealed class ClientPSHost
    {
        /// <summary>
        /// Constrctor of ClientPSHost
        /// </summary>
        /// <param name="host">IPEndPoint of Host (default installation 127.0.0.1:8080)</param>
        /// <param name="UI">Local PSHostUserInterface(If use in PowerShell.exe, it would be $host.UI)</param>
        /// <param name="channel">1: NegotiateChannel;  2: SSL Channel </param>
        /// <param name="Target"> Certificate Name of server, usually full qulified DNS name of server </param>
        /// <remarks>
        /// PSHostRawUserInterface (PSHostUserInterface.RawUI) is not mandatory.   
        /// </remarks>
        /// <exception cref="ArgumentNullException">
        /// host is null
        /// -or-
        /// UI is null
        /// </exception>
        /// <example>
        /// static void Main(string[] args)
        /// {
        ///      System.Net.IPEndPoint RemoteEP = new System.Net.IPEndPoint(System.Net.IPAddress.Parse("127.0.0.1"), 8080);
        ///      System.Management.Automation.Host.PSHostUserInterface LocalUI = new LocalPSHostUserInterface();
        ///      ClientPSHost clientHost = new ClientPSHost(RemoteEP, LocalUI);
        ///      clientHost.Start();
        /// }
        /// </example>
        public ClientPSHost(IPEndPoint host, PSHostUserInterface UI, int channel, string Target)
        {
            if (host == null)
                throw new ArgumentNullException("IPAddress");
            this.HostEP = host;
            if (UI == null)
                throw new ArgumentNullException("UI");
            if (channel < 1 || channel > 2)
                throw new ArgumentOutOfRangeException("channel");
            if (channel == 2 && string.IsNullOrEmpty(Target))
                throw new ArgumentNullException("Target");
            this.OutterUI = UI;
            if (UI.RawUI == null)
                this.HasRawUI = false;
            else
                this.HasRawUI = true;
            this.version = new Version(0, 2, 9, 0);
            this.instanceId = Guid.NewGuid();
            this.Channel = channel;
            this.TargetName = Target;

        }

        private const int NormalExit = 0;
        private const int ServiceShutDown = 1;
        private const int ClientShutDown = 2;
        private const int MaxClient = 3;
        private const int MaxClientPerIP = 4;
        private const int ErrorChannelInitialization = 100;
        private const int ErrorChannelSocket = 101;
        private const int ErrorChannelOther = 102;
        private const int ErrorRunspaceInitialization = 200;
        private const int ErrorRunspace = 201;
        private const int ErrorUI = 301;
        private const int ErrorRawUI = 302;
        private const int ErrorOtherGeneral = 401;


        private bool ShouldExit = false;

        private int Channel = 0;

        private string TargetName = string.Empty;

        private bool canThrowExcexption = false;
        /// <summary>
        /// Get or Set the value indicating whether ClientPSHost Can Throw Exception 
        /// </summary>
        /// <remarks>
        /// Default value is false. It is recommended that set this property to false.
        /// If set to false, ClientPSHost will use local PSHostRawUserInterface to report exceptions
        /// If set this property to true, outter application should catch and process exception.
        /// </remarks>
        /// <example>
        /// static void Main(string[] args)
        /// {
        ///      System.Net.IPEndPoint RemoteEP = new System.Net.IPEndPoint(System.Net.IPAddress.Parse("127.0.0.1"), 8080);
        ///      System.Management.Automation.Host.PSHostUserInterface LocalUI = new LocalPSHostUserInterface();
        ///      ClientPSHost clientHost = new ClientPSHost(RemoteEP, LocalUI);
        ///      clientHost.CanThrowException = true;
        ///      clientHost.Start();
        /// }
        /// </example>
        public bool CanThrowException
        {
            get { return this.canThrowExcexption; }
            set { this.canThrowExcexption = value; }
        }

        private bool handleCancelKey = true;
        /// <summary>
        /// Get or Set the value indicating whether ClientPSHost Can Handle CancelKey (Ctrl+C and Ctrl+Break) 
        /// </summary>
        /// <remarks>
        /// Default value is true. It is recommended that set this property to true.
        /// If set to true, ClientPSHost will handle Console.CancelKeyPress Event. 
        /// If set this property to false, ClientPSHost will not send Console.CancelKeyPress Event to remote host.
        /// </remarks>
        /// <example>
        /// static void Main(string[] args)
        /// {
        ///      System.Net.IPEndPoint RemoteEP = new System.Net.IPEndPoint(System.Net.IPAddress.Parse("127.0.0.1"), 8080);
        ///      System.Management.Automation.Host.PSHostUserInterface LocalUI = new LocalPSHostUserInterface();
        ///      ClientPSHost clientHost = new ClientPSHost(RemoteEP, LocalUI);
        ///      clientHost.CanHandleCancelKey = false;
        ///      clientHost.Start();
        /// }
        /// </example>
        public bool CanHandleCancelKey
        {
            get { return this.handleCancelKey; }
            set { this.handleCancelKey = value; }
        }
        
        private int ExitCode = 0;

        private string OldWindowTitle;

        private int OldCursorSize = 25;

        private Size OldWindowSize = new Size();

        private Coordinates OldWindowPosition = new Coordinates();

        private ConsoleColor OldForeColor;

        private ConsoleColor OldBackColor;

        private bool OldTreatControlC;

        private bool SettingSaved = false;

        private IPEndPoint HostEP = null;

        private PSHostUserInterface OutterUI = null;

        private SecureChannel MyChannel = null;

        private TcpClient Client = null;

        private UdpClient CtrlSender = null;

        private NetworkStream InnerStream = null;

        private bool HasRawUI = false;

        private CultureInfo originalCultureInfo = System.Threading.Thread.CurrentThread.CurrentCulture;

        private CultureInfo originalUICultureInfo = System.Threading.Thread.CurrentThread.CurrentUICulture;

        private Guid instanceId;
        /// <summary>
        /// Get a Guid object represent current unique instance id of ClientPSHost 
        /// </summary>
        public Guid InstanceId
        {
            get { return instanceId; }
        }

        /// <summary>
        /// Always return "ClientPSHost"
        /// </summary>
        public string Name
        {
            get { return "ClientPSHost"; }
        }

        /// <summary>
        /// Get PSHostUserInterface object used by current ClientPSHost
        /// </summary>
        public PSHostUserInterface UI
        {
            get { return this.OutterUI; }
        }

        /// <summary>
        /// Get current version of ClientPSHost
        /// </summary>
        public Version Version
        {
            get { return this.version; }
        }
        private Version version;

        /// <summary>
        /// Start ClientPSHost
        /// </summary>
        /// <remarks>
        /// ClientPSHost will connect to server, ask for login credential, initialize a secure channel and start a remote host.
        /// </remarks>
        public void Start()
        {
            this.Initialize();
            this.ListenLoop();
            this.Close();
        }

        private void ReportException(Exception ex)
        {
            if (ex != null)
            {
                try
                {
                    this.OutterUI.WriteErrorLine(ex.Message);
#if ClientDebug
                    this.OutterUI.WriteErrorLine(ex.Source);
                    this.OutterUI.WriteErrorLine(ex.StackTrace);
#endif
                    this.OutterUI.WriteLine();
                }
                catch
                {
                    if (this.canThrowExcexption)
                        throw;
                }
            }
        }
        
        /// <summary>
        /// Force ClientPSHost to disconnect from remote host and exit
        /// </summary>
        /// <param name="exitCode">Exit code of ClientPSHost</param>
        public void SetShouldExit(int exitCode)
        {
            this.ShouldExit = true;
            this.ExitCode = exitCode;
            try
            {
                this.OutterUI.WriteLine(ConsoleColor.White, ConsoleColor.Black, "PowerShell Remoting Client Quiting ...");
                this.OutterUI.WriteLine(ConsoleColor.White, ConsoleColor.Black, "Exitcode: " + exitCode.ToString());
                this.OutterUI.WriteLine();
            }
            catch 
            {
                if (this.canThrowExcexption)
                    throw;                
            }
        }

        private void HandShake(TcpClient client, ref string Message)
        {
            if (client == null)
                throw new ArgumentNullException("client");
            if (string.IsNullOrEmpty(Message))
                throw new ArgumentNullException("Message");
            NetworkStream stream = null;
            byte[] MessageBytes = new byte[1024];
            int length;
            string IncomeMessage = string.Empty;
            stream = client.GetStream();
            stream.ReadTimeout = 5000;
            stream.WriteTimeout = 5000;
            MessageBytes = Encoding.ASCII.GetBytes(Message);
            stream.Write(MessageBytes, 0, MessageBytes.Length);
            length = stream.Read(MessageBytes, 0, MessageBytes.Length);
            IncomeMessage = Encoding.ASCII.GetString(MessageBytes, 0, length);
            stream.ReadTimeout = System.Threading.Timeout.Infinite;
            stream.WriteTimeout = System.Threading.Timeout.Infinite;
            Message = IncomeMessage;

        }

        private void Initialize()
        {
            try
            {
                this.Welcome();
                this.SaveOldSetting();
                this.OutterUI.WriteLine("Connecting to " + this.HostEP.Address.ToString() + " : " + this.HostEP.Port.ToString());
                this.Client = new TcpClient();
                this.Client.Connect(this.HostEP);
                string Message;
                if (this.Channel == 2)
                    Message = "~SSLChannel";
                else
                    Message = "~NegotiateChannel";
                this.HandShake(this.Client, ref Message);
                if (Message == "~MaxClient")
                {
                    this.OutterUI.WriteErrorLine("Server reach its maximum client capacity! Try again later.");
                    this.OutterUI.WriteLine();
                    this.SetShouldExit(ClientPSHost.MaxClient);
                    return;
                }
                else if (Message == "~MaxClientPerIP")
                {
                    this.OutterUI.WriteErrorLine("Too many conection from same IP! Close unused client and try again.");
                    this.OutterUI.WriteLine();
                    this.SetShouldExit(ClientPSHost.MaxClientPerIP);
                    return;
                }
                this.InnerStream = Client.GetStream();
                if (this.Channel == 2)
                {
                    Message = "~SSLChannel";
                    this.HandShake(this.Client, ref Message);
                    if (Message == "~NoSupport")
                    {
                        this.OutterUI.WriteErrorLine("SSLChannel is not supported by server.");
                        this.OutterUI.WriteLine();
                        this.SetShouldExit(ClientPSHost.ErrorChannelOther);
                        return;
                    }
                    else if (Message != "~Welcome")
                    {
                        this.OutterUI.WriteErrorLine("Unexpected response: " + Message);
                        this.OutterUI.WriteLine();
                        this.SetShouldExit(ClientPSHost.ErrorChannelOther);
                        return;
                    }
                    this.OutterUI.WriteLine("Initializing SSL Channel...");
                    this.OutterUI.WriteLine();
                    this.MyChannel = new SSLChannel(this.InnerStream, null);
                }
                else
                {
                    this.OutterUI.WriteLine("Initializing Negotiate Channel...");
                    this.MyChannel = new NegotiateChannel(this.InnerStream);
                }
                NetworkCredential LoginCre = GetCredential();
                if (LoginCre == null)
                {
                    this.OutterUI.WriteErrorLine("Null credential");
                    this.SetShouldExit(ClientPSHost.ErrorChannelInitialization);
                    return;
                }
                this.MyChannel.OpenAsClient(LoginCre, this.TargetName);
                if (!this.MyChannel.IsAuthenticated)
                {
                    this.OutterUI.WriteErrorLine("Authentication error!");
                    this.SetShouldExit(ClientPSHost.ErrorChannelInitialization);
                    return;
                }
                if (this.Channel == 2)                
                {
                    if (this.MyChannel.ReceiveCommand() != SecureChannel.OK)
                    {
                        this.OutterUI.WriteErrorLine("Unexpected server response!");
                        this.SetShouldExit(ClientPSHost.ErrorChannelInitialization);
                        return;
                    }
                    X509Certificate remoteCertificate = ((SSLChannel)this.MyChannel).RemoteCertificate;
                    if (remoteCertificate == null)
                    {
                        this.OutterUI.WriteErrorLine("Can't get remote X509 certificate!");
                        this.MyChannel.SendCommand(SecureChannel.Cancal);
                        this.SetShouldExit(ClientPSHost.ErrorChannelInitialization);
                        return;
                    }
                    DateTime ConnectionDate = DateTime.Now;
                    DateTime EffectiveDate = DateTime.Parse(remoteCertificate.GetEffectiveDateString());
                    DateTime ExpirationDate = DateTime.Parse(remoteCertificate.GetExpirationDateString());
                    if (ConnectionDate < EffectiveDate || ConnectionDate > ExpirationDate)
                    {
                        this.OutterUI.WriteErrorLine("X509 certificate expirated!");
                        this.MyChannel.SendCommand(SecureChannel.Cancal);
                        this.SetShouldExit(ClientPSHost.ErrorChannelInitialization);
                        return;
                    }
                    this.OutterUI.WriteLine(string.Format("Remote Server: {0}",remoteCertificate.Subject));
                    this.OutterUI.WriteLine(string.Format("Issued by: {0}", remoteCertificate.Issuer));
                    this.OutterUI.WriteLine(string.Format("Effective Date: {0}", remoteCertificate.GetEffectiveDateString()));
                    this.OutterUI.WriteLine(string.Format("Expiration Date: {0}", remoteCertificate.GetExpirationDateString()));
                    this.OutterUI.WriteLine(string.Format("Serial Number: {0}", remoteCertificate.GetSerialNumberString()));
                    this.OutterUI.WriteLine(string.Format("Certificate Hash: {0}", remoteCertificate.GetCertHashString()));
                    Collection<ChoiceDescription> AcceptCertificate = new Collection<ChoiceDescription>();
                    ChoiceDescription Yes = new ChoiceDescription("&Yes", "Accept this certificate and connect to server");
                    ChoiceDescription No = new ChoiceDescription("&No", "Reject this certificate and disconnect to server");
                    AcceptCertificate.Add(Yes);
                    AcceptCertificate.Add(No);
                    this.OutterUI.WriteLine();
                    int choice = this.OutterUI.PromptForChoice("", "Accept this certificate and connect to server?", AcceptCertificate, 0);
                    this.OutterUI.WriteLine();
                    if (choice < 0 || choice >= 1)
                    {
                        this.MyChannel.SendCommand(SecureChannel.Cancal);
                        this.SetShouldExit(ClientPSHost.ErrorChannelInitialization);
                        return;
                    }
                    this.MyChannel.SendCommand(SecureChannel.OK);
                }
                this.OutterUI.WriteLine("OK");
                int command = this.MyChannel.ReceiveCommand();
                if (command < IPEndPoint.MaxPort && command > IPEndPoint.MinPort && this.handleCancelKey)
                {
                    this.CtrlSender = new UdpClient(new IPEndPoint(((IPEndPoint)(this.Client.Client.LocalEndPoint)).Address,0));
                    this.MyChannel.SendCommand(((IPEndPoint)(this.CtrlSender.Client.LocalEndPoint)).Port);
                    IPEndPoint HostUdpEP = new IPEndPoint(this.HostEP.Address, command);
                    this.CtrlSender.Connect(HostUdpEP);
                    ConsoleCancelEventHandler CtrlHandler = new ConsoleCancelEventHandler(this.CtrlPressed);
                    Console.CancelKeyPress += CtrlHandler;
                    Console.TreatControlCAsInput = false;
                }
                else
                {
                    this.MyChannel.SendCommand(-1);
                }

            }
            catch (Exception ex)
            {
                this.ReportException(ex);
                this.SetShouldExit(ClientPSHost.ErrorChannelOther);
                if (this.canThrowExcexption)
                    throw;
            }
        }

        private void Welcome()
        {
            this.OutterUI.WriteLine();
            this.OutterUI.WriteLine(ConsoleColor.White,ConsoleColor.Black,"PowerShell Remoting Client " + this.version.ToString());
            this.OutterUI.WriteLine(ConsoleColor.White, ConsoleColor.Black, "by http://MSHForFun.blogspot.com/ 2006");
            this.OutterUI.WriteLine();
        }
        
        private NetworkCredential GetCredential()
        {
            PSCredential LoginCre;
            if (this.Channel == 2)
            {
                LoginCre = this.OutterUI.PromptForCredential("", "Please provide login credential", "", "");
                if (LoginCre != null)
                    return LoginCre.GetNetworkCredential();
                else
                    return null;
            }
            else
            {
                Collection<ChoiceDescription> UseDefault = new Collection<ChoiceDescription>();
                ChoiceDescription Yes = new ChoiceDescription("&Yes", "Login as current user");
                ChoiceDescription No = new ChoiceDescription("&No", "Login as different user");
                UseDefault.Add(Yes);
                UseDefault.Add(No);
                int choice = this.OutterUI.PromptForChoice("", "Use default credential?", UseDefault, 0);
                this.OutterUI.WriteLine();
                if (choice <= 0 || choice >= 2)
                    return CredentialCache.DefaultNetworkCredentials;
                LoginCre = this.OutterUI.PromptForCredential("", "Please provide login credential", "", "");
                if (LoginCre != null)
                    return LoginCre.GetNetworkCredential();
                else
                    return null;
            }
        }

        private void Close()
        {
            if (this.Client != null)
                this.Client.Close();
            if (this.MyChannel != null)
                this.MyChannel.Close();
            if (this.InnerStream != null)
                this.InnerStream.Close();
            this.Client = null;
            this.MyChannel = null;
            this.InnerStream = null;
        }

        private void SaveOldSetting()
        {
            if (this.handleCancelKey)
                this.OldTreatControlC = Console.TreatControlCAsInput;
            if (!this.ShouldExit && this.HasRawUI)
            {
                this.OldBackColor = this.UI.RawUI.BackgroundColor;
                this.OldCursorSize = this.UI.RawUI.CursorSize;
                this.OldForeColor = this.UI.RawUI.ForegroundColor;
                this.OldWindowPosition = this.UI.RawUI.WindowPosition;
                this.OldWindowSize = this.UI.RawUI.WindowSize;
                this.OldWindowTitle = this.UI.RawUI.WindowTitle;
                this.SettingSaved = true;
            }
            else
            {
                this.SettingSaved = false;
            }
        }

        private void RestoreOldSetting()
        {
            if (this.handleCancelKey)
            {
                Console.TreatControlCAsInput = this.OldTreatControlC;
                ConsoleCancelEventHandler CtrlHandler = new ConsoleCancelEventHandler(this.CtrlPressed);
                Console.CancelKeyPress -= CtrlHandler;
            }
            if (this.HasRawUI && this.SettingSaved)
            {
                if (this.UI.RawUI.BackgroundColor != this.OldBackColor)
                    this.UI.RawUI.BackgroundColor = this.OldBackColor;
                if (this.UI.RawUI.CursorSize != this.OldCursorSize)
                    this.UI.RawUI.CursorSize = this.OldCursorSize;
                if (this.UI.RawUI.ForegroundColor != this.OldForeColor)
                    this.UI.RawUI.ForegroundColor = this.OldForeColor;
                if (this.UI.RawUI.WindowPosition != this.OldWindowPosition)
                    this.UI.RawUI.WindowPosition = this.OldWindowPosition;
                if (this.UI.RawUI.WindowSize != this.OldWindowSize)
                    this.UI.RawUI.WindowSize = this.OldWindowSize;
                if (this.UI.RawUI.WindowTitle != this.OldWindowTitle)
                    this.UI.RawUI.WindowTitle = this.OldWindowTitle;
            }
        }

        private void CtrlPressed(Object sender,ConsoleCancelEventArgs e)
        {
            byte[] data;
            if (e.SpecialKey == ConsoleSpecialKey.ControlC)
            {
                data = BitConverter.GetBytes(-100); 
                this.CtrlSender.Send(data,data.Length);
                e.Cancel = true;
            }
            else if (e.SpecialKey == ConsoleSpecialKey.ControlBreak)
            {
                data = BitConverter.GetBytes(-200);
                this.CtrlSender.Send(data, data.Length);
            }
 
        }

        private void ListenLoop()
        {
            int Command = 0;
            try
            {
                while (!this.ShouldExit)
                {
                    Command = this.MyChannel.ReceiveCommand();
                    switch (Command)
                    {
                        case SecureChannel.ReadLine:
                            this.ReadLine();
                            break;
                        case SecureChannel.Write:
                            this.Write();
                            break;
                        case SecureChannel.ReadLineAsSecureString:
                            this.ReadLineAsSecureString();
                            break;
                        case SecureChannel.PromptForChoice:
                            this.PromptForChoice();
                            break;
                        case SecureChannel.PromptForCredential:
                            this.PromptForCredential();
                            break;
                        case SecureChannel.WriteColor:
                            this.WriteColor();
                            break;
                        case SecureChannel.WriteLine:
                            this.WriteLine();
                            break;
                        case SecureChannel.WriteLineColor:
                            this.WriteLineColor();
                            break;
                        case SecureChannel.WriteEmptyLine:
                            this.WriteEmptyLine();
                            break;
                        case SecureChannel.WriteErrorLine:
                            this.WriteErrorLine();
                            break;
                        case SecureChannel.WriteVerboseLine:
                            this.WriteVerboseLine();
                            break;
                        case SecureChannel.WriteDebugLine:
                            this.WriteDebugLine();
                            break;
                        case SecureChannel.WriteProgress:
                            this.WriteProgress();
                            break;
                        case SecureChannel.RawGetBackgroundColor:
                            this.GetBackgroundColor();
                            break;
                        case SecureChannel.RawSetBackgroundColor:
                            this.SetBackgroundColor();
                            break;
                        case SecureChannel.RawGetForegroundColor:
                            this.GetForegroundColor();
                            break;
                        case SecureChannel.RawSetForegroundColor:
                            this.SetForegroundColor();
                            break;
                        case SecureChannel.RawGetBufferSize:
                            this.GetBufferSize();
                            break;
                        case SecureChannel.RawSetBufferSize:
                            this.SetBufferSize();
                            break;
                        case SecureChannel.RawGetCursorPosition:
                            this.GetCursorPosition();
                            break;
                        case SecureChannel.RawSetCursorPosition:
                            this.SetCursorPosition();
                            break;
                        case SecureChannel.RawGetCursorSize:
                            this.GetCursorSize();
                            break;
                        case SecureChannel.RawSetCursorSize:
                            this.SetCursorSize();
                            break;
                        case SecureChannel.RawKeyAvailable:
                            this.KeyAvailable();
                            break;
                        case SecureChannel.RawGetMaxPhysicalWindowSize:
                            this.GetMaxPhysicalWindowSize();
                            break;
                        case SecureChannel.RawGetMaxWindowSize:
                            this.GetMaxWindowSize();
                            break;
                        case SecureChannel.RawGetWindowPosition:
                            this.GetWindowPosition();
                            break;
                        case SecureChannel.RawSetWindowPosition:
                            this.SetWindowPosition();
                            break;
                        case SecureChannel.RawGetWindowSize:
                            this.GetWindowSize();
                            break;
                        case SecureChannel.RawSetWindowSize:
                            this.SetWindowSize();
                            break;
                        case SecureChannel.RawGetWindowTitle:
                            this.GetWindowTitle();
                            break;
                        case SecureChannel.RawSetWindowTitle:
                            this.SetWindowTitle();
                            break;
                        case SecureChannel.RawReadKey:
                            this.ReadKey();
                            break;
                        case SecureChannel.RawFlushInputBuffer:
                            this.FlushInputBuffer();
                            break;
                        case SecureChannel.RawScrollBufferContents:
                            this.ScrollBufferContents();
                            break;
                        case SecureChannel.RawSetBufferContents1:
                            this.SetBufferContents1();
                            break;
                        case SecureChannel.RawSetBufferContents2:
                            this.SetBufferContents2();
                            break;
                        case SecureChannel.RawGetBufferContents:
                            this.GetBufferContents();
                            break;
                        case SecureChannel.Exit:
                            System.Threading.Thread.Sleep(1000);
                            this.SetShouldExit(0);
                            return;
                        default:
                            this.MyChannel.SendCommand(SecureChannel.Error);
                            break;
                    }
                }
                this.RestoreOldSetting();
            }
            catch (Exception ex)
            {
                this.ReportException(ex);
                this.SetShouldExit(200);
                if (this.canThrowExcexption)
                    throw;
            }

        }

        private void ReadLine()
        {
            string line;
            try
            {
                line = this.OutterUI.ReadLine();
            }
            catch (Exception ex)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            if (string.IsNullOrEmpty(line))
                this.MyChannel.WriteText(SecureChannel.EmptyLine);
            else
                this.MyChannel.WriteText(line);
        }

        private void Write()
        {
            this.MyChannel.SendCommand(SecureChannel.OK);
            string line = this.MyChannel.ReadText();
            if (!string.IsNullOrEmpty(line))
            {
                try
                {
                    this.OutterUI.Write(line);
                }
                catch (Exception ex)
                {
                    this.MyChannel.SendCommand(SecureChannel.Error);
                    if (this.canThrowExcexption)
                        throw;
                    else
                        this.OutterUI.WriteErrorLine(ex.Message);
                    return;
                }
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void WriteColor()
        {
            this.MyChannel.SendCommand(SecureChannel.OK);
            string line = this.MyChannel.ReadText();
            string[] split = line.Split(new Char[] { '~' });
            if (split.Length != 3)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                throw new ArgumentOutOfRangeException("WriteColor");
            }
            if (!string.IsNullOrEmpty(split[2]))
            {
                ConsoleColor Fore = (ConsoleColor)(Int32.Parse(split[0]));
                ConsoleColor Back = (ConsoleColor)(Int32.Parse(split[1]));
                try
                {
                    this.OutterUI.Write(Fore, Back, split[2]);
                }
                catch (Exception ex)
                {
                    this.MyChannel.SendCommand(SecureChannel.Error);
                    if (this.canThrowExcexption)
                        throw;
                    else
                        this.OutterUI.WriteErrorLine(ex.Message);
                    return;
                }
            }
            this.MyChannel.SendCommand(SecureChannel.OK); 
        }

        private void WriteLineColor()
        {
            this.MyChannel.SendCommand(SecureChannel.OK);
            string line = this.MyChannel.ReadText();
            string[] split = line.Split(new Char[] { '~' });
            if (split.Length != 3)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                throw new ArgumentOutOfRangeException("WriteLineColor");
            }
            if (!string.IsNullOrEmpty(split[2]))
            {
                ConsoleColor Fore = (ConsoleColor)(Int32.Parse(split[0]));
                ConsoleColor Back = (ConsoleColor)(Int32.Parse(split[1]));
                try
                {
                    this.OutterUI.WriteLine(Fore, Back, split[2]);
                }
                catch (Exception ex)
                {
                    this.MyChannel.SendCommand(SecureChannel.Error);
                    if (this.canThrowExcexption)
                        throw;
                    else
                        this.OutterUI.WriteErrorLine(ex.Message);
                    return;
                }
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        
        private void WriteEmptyLine()
        {
            try
            {
                this.OutterUI.WriteLine();
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void WriteLine()
        {
            this.MyChannel.SendCommand(SecureChannel.OK);
            string line = this.MyChannel.ReadText();
            if (!string.IsNullOrEmpty(line))
            {
                try
                {
                    this.OutterUI.WriteLine(line);
                }
                catch (Exception ex)
                {
                    this.MyChannel.SendCommand(SecureChannel.Error);
                    if (this.canThrowExcexption)
                        throw;
                    else
                        this.OutterUI.WriteErrorLine(ex.Message);
                    return;
                }
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void WriteErrorLine()
        {
            this.MyChannel.SendCommand(SecureChannel.OK);
            string line = this.MyChannel.ReadText();
            if (!string.IsNullOrEmpty(line))
            {
                try
                {
                    this.OutterUI.WriteErrorLine(line);
                }
                catch (Exception ex)
                {
                    this.MyChannel.SendCommand(SecureChannel.Error);
                    if (this.canThrowExcexption)
                        throw;
                    else
                        this.OutterUI.WriteErrorLine(ex.Message);
                    return;
                }
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void WriteVerboseLine()
        {
            this.MyChannel.SendCommand(SecureChannel.OK);
            string line = this.MyChannel.ReadText();
            if (!string.IsNullOrEmpty(line))
            {
                try
                {
                    this.OutterUI.WriteVerboseLine(line);
                }
                catch (Exception ex)
                {
                    this.MyChannel.SendCommand(SecureChannel.Error);
                    if (this.canThrowExcexption)
                        throw;
                    else
                        this.OutterUI.WriteErrorLine(ex.Message);
                    return;
                }
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void WriteDebugLine()
        {
            this.MyChannel.SendCommand(SecureChannel.OK);
            string line = this.MyChannel.ReadText();
            if (!string.IsNullOrEmpty(line))
            {
                try
                {
                    this.OutterUI.WriteDebugLine(line);
                }
                catch (Exception ex)
                {
                    this.MyChannel.SendCommand(SecureChannel.Error);
                    if (this.canThrowExcexption)
                        throw;
                    else
                        this.OutterUI.WriteErrorLine(ex.Message);
                    return;
                }
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void ReadLineAsSecureString()
        {
            byte[] Buffer = new byte[1];
            SecureString SS = new SecureString();
            try
            {
                SS = this.OutterUI.ReadLineAsSecureString();
            }
            catch (Exception ex)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            if (SS.Length == 0)
            {
                this.MyChannel.WriteText(SecureChannel.EmptyLine);
            }
            this.MyChannel.WriteText(SecureStringHelper.GetStringFromSecureString(SS));
        }

        private void GetBackgroundColor()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            ConsoleColor Color;
            try
            {
                Color = this.OutterUI.RawUI.BackgroundColor;
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand((int)Color);
        }

        private void SetBackgroundColor()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI"); 

                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
            int Command = this.MyChannel.ReceiveCommand();
            try
            {
                this.OutterUI.RawUI.BackgroundColor = (ConsoleColor)Command;
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void GetForegroundColor()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI"); 
                return;
            }
            ConsoleColor Color;
            try
            {
                Color = this.OutterUI.RawUI.ForegroundColor;
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand((int)Color);
        }

        private void SetForegroundColor()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI"); 
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
            int Command = this.MyChannel.ReceiveCommand();
            try
            {
                this.OutterUI.RawUI.ForegroundColor = (ConsoleColor)Command;
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void GetBufferSize()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI"); 
                return;
            }
            Size buffer;
            try
            {
                buffer = this.OutterUI.RawUI.BufferSize;
            }
            catch (Exception ex)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.WriteText(buffer.Width.ToString() + "~" + buffer.Height.ToString());
        }

        private void SetBufferSize()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI"); 
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
            string data = this.MyChannel.ReadText();
            string[] split = data.Split(new Char[] { '~' });
            if (split.Length != 2)
                throw new ArgumentOutOfRangeException("split");
            try
            {
                Size buffer = new Size(int.Parse(split[0]),int.Parse(split[1]));
                this.OutterUI.RawUI.BufferSize = buffer;
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void GetCursorPosition()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI"); 
                return;
            }
            Coordinates cursor;
            try
            {
                cursor = this.OutterUI.RawUI.CursorPosition;
            }
            catch (Exception ex)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.WriteText(cursor.X.ToString() + "~" + cursor.Y.ToString());
        }

        private void SetCursorPosition()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
            string data = this.MyChannel.ReadText();
            string[] split = data.Split(new Char[] { '~' });
            if (split.Length != 2)
                throw new ArgumentOutOfRangeException("split");
            try
            {
                Coordinates cursor = new  Coordinates(Int32.Parse(split[0]), Int32.Parse(split[1]));
                this.OutterUI.RawUI.CursorPosition = cursor;
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void GetCursorSize()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            int cursor;
            try
            {
                cursor = this.OutterUI.RawUI.CursorSize;
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand(cursor);
        }

        private void SetCursorSize()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
            int Command = this.MyChannel.ReceiveCommand();
            try
            {
                this.OutterUI.RawUI.CursorSize = Command;
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void KeyAvailable()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            bool keyavailable;
            try
            {
                keyavailable = this.OutterUI.RawUI.KeyAvailable;
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            if (keyavailable)
                this.MyChannel.SendCommand(1);
            else
                this.MyChannel.SendCommand(0);
        }

        private void GetMaxPhysicalWindowSize()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI"); 
                return;
            }
            Size window;
            try
            {
                window = this.OutterUI.RawUI.MaxPhysicalWindowSize;
            }
            catch (Exception ex)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.WriteText(window.Width.ToString() + "~" + window.Height.ToString());
        }

        private void GetMaxWindowSize()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            Size window;
            try
            {
                window = this.OutterUI.RawUI.MaxWindowSize;
            }
            catch (Exception ex)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.WriteText(window.Width.ToString() + "~" + window.Height.ToString());
        }

        private void GetWindowPosition()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            Coordinates window;
            try
            {
                window = this.OutterUI.RawUI.WindowPosition;
            }
            catch (Exception ex)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.WriteText(window.X.ToString() + "~" + window.Y.ToString());
        }

        private void SetWindowPosition()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
            string data = this.MyChannel.ReadText();
            string[] split = data.Split(new Char[] { '~' });
            if (split.Length != 2)
                throw new ArgumentOutOfRangeException("split");
            try
            {
                Coordinates window = new Coordinates(Int32.Parse(split[0]), Int32.Parse(split[1]));
                this.OutterUI.RawUI.WindowPosition = window;
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void GetWindowSize()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            Size window;
            try
            {
                window = this.OutterUI.RawUI.WindowSize;
            }
            catch (Exception ex)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.WriteText(window.Width.ToString() + "~" + window.Height.ToString());
        }

        private void SetWindowSize()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
            string data = this.MyChannel.ReadText();
            string[] split = data.Split(new Char[] { '~' });
            if (split.Length != 2)
                throw new ArgumentOutOfRangeException("split");
            try
            {
                Size window = new Size(Int32.Parse(split[0]), Int32.Parse(split[1]));
                this.OutterUI.RawUI.WindowSize = window;
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void GetWindowTitle()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            string window;
            try
            {
                window = this.OutterUI.RawUI.WindowTitle;
            }
            catch (Exception ex)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            if (string.IsNullOrEmpty(window))
                this.MyChannel.WriteText(SecureChannel.EmptyLine);
            else
                this.MyChannel.WriteText(window);
        }

        private void SetWindowTitle()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
            string data = this.MyChannel.ReadText();
            if (data == SecureChannel.EmptyLine)
                data = string.Empty;
            try
            {
                this.OutterUI.RawUI.WindowTitle = data;
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void ReadKey()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
            int Command = this.MyChannel.ReceiveCommand();
            KeyInfo Key = new KeyInfo();
            try
            {
                ReadKeyOptions options=  (ReadKeyOptions) Command;
                Key = this.OutterUI.RawUI.ReadKey(options);
            }
            catch (Exception ex)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            StringBuilder sb = new StringBuilder();
            sb.Append(Key.Character);
            sb.Append("~"+((int)Key.ControlKeyState).ToString());
            sb.Append("~"+Key.VirtualKeyCode.ToString());
            if (Key.KeyDown)
                sb.Append("~Y");
            else
                sb.Append("~N");
            this.MyChannel.WriteText(sb.ToString());
        }

        private void FlushInputBuffer()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            try
            {
                this.OutterUI.RawUI.FlushInputBuffer();
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void ScrollBufferContents()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
            string data = this.MyChannel.ReadText();
            string[] split = data.Split(new Char[] { '~' });
            if (split.Length != 14)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Can't parse parameters");
                return;
            }
            Rectangle source = new Rectangle();
            Coordinates destination = new Coordinates();
            Rectangle clip = new Rectangle();
            BufferCell fill= new BufferCell();
            source.Top = int.Parse(split[0]);
            source.Bottom = int.Parse(split[1]);
            source.Left = int.Parse(split[2]);
            source.Right = int.Parse(split[3]);
            destination.X = int.Parse(split[4]);
            destination.Y = int.Parse(split[5]);
            clip.Top = int.Parse(split[6]);
            clip.Bottom = int.Parse(split[7]);
            clip.Left = int.Parse(split[8]);
            clip.Right = int.Parse(split[9]);
            fill.Character = split[10][0];
            fill.ForegroundColor = (ConsoleColor)(int.Parse(split[11]));
            fill.BackgroundColor = (ConsoleColor)(int.Parse(split[12]));
            fill.BufferCellType = (BufferCellType)(int.Parse(split[13]));
            try
            {
                this.OutterUI.RawUI.ScrollBufferContents(source, destination, clip, fill);
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void SetBufferContents1()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
            string data = this.MyChannel.ReadText();
            string[] split = data.Split(new Char[] { '~' });
            if (split.Length != 8)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Can't parse parameters");
                return;
            }
            Rectangle source = new Rectangle();
            BufferCell fill = new BufferCell();
            source.Top = int.Parse(split[0]);
            source.Bottom = int.Parse(split[1]);
            source.Left = int.Parse(split[2]);
            source.Right = int.Parse(split[3]);
            fill.Character = split[4][0];
            fill.ForegroundColor = (ConsoleColor)(int.Parse(split[5]));
            fill.BackgroundColor = (ConsoleColor)(int.Parse(split[6]));
            fill.BufferCellType = (BufferCellType)(int.Parse(split[7]));
            try
            {
                this.OutterUI.RawUI.SetBufferContents(source, fill);
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void GetBufferContents()
        {
            byte[] errordata = new byte[1];
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
            string data = this.MyChannel.ReadText();
            string[] split = data.Split(new Char[] { '~' });
            if (split.Length != 4)
            {
                this.MyChannel.WriteData(errordata, 0, 1);
                this.OutterUI.WriteErrorLine("RawUI Error: Can't parse parameters");
                return;
            }
            Rectangle source = new Rectangle();
            source.Top = int.Parse(split[0]);
            source.Bottom = int.Parse(split[1]);
            source.Left = int.Parse(split[2]);
            source.Right = int.Parse(split[3]);
            BufferCell[,] results = null;
            try
            {
                results = this.OutterUI.RawUI.GetBufferContents(source);
            }
            catch (Exception ex)
            {
                this.MyChannel.WriteData(errordata, 0, 1);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
            int command = this.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
            {
                this.MyChannel.WriteData(errordata,0,1);
                this.OutterUI.WriteErrorLine("RawUI Error: unexpected resonse from server");
                return;
            }
            int Width = source.Right - source.Left + 1;
            int Highth = source.Bottom - source.Top + 1;
            int BufferCellSize = 14; // sizeof(char) + sizeof(int) * 3;
            byte[] RawData = new byte[Width * Highth * BufferCellSize];
            int k = 0;
            byte[] tmpdata = null; 
            for (int i = 0; i < Highth; i++)
            {
                for (int j = 0; j < Width; j++)
                {
                    tmpdata = BitConverter.GetBytes(results[i, j].Character);
                    RawData[k] = tmpdata[0];
                    RawData[k+1] = tmpdata[1];
                    k += 2;
                    tmpdata = BitConverter.GetBytes(((int)results[i, j].ForegroundColor));
                    RawData[k] = tmpdata[0];
                    RawData[k + 1] = tmpdata[1];
                    RawData[k + 2] = tmpdata[2];
                    RawData[k + 3] = tmpdata[3];
                    k += 4;
                    tmpdata = BitConverter.GetBytes(((int)results[i, j].BackgroundColor));
                    RawData[k] = tmpdata[0];
                    RawData[k + 1] = tmpdata[1];
                    RawData[k + 2] = tmpdata[2];
                    RawData[k + 3] = tmpdata[3];
                    k += 4;
                    tmpdata = BitConverter.GetBytes(((int)results[i, j].BufferCellType));
                    RawData[k] = tmpdata[0];
                    RawData[k + 1] = tmpdata[1];
                    RawData[k + 2] = tmpdata[2];
                    RawData[k + 3] = tmpdata[3];
                    k += 4;
                }
            }
            this.MyChannel.WriteData(RawData,0,RawData.Length);
        }

        private void SetBufferContents2()
        {
            if (!this.HasRawUI)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Null RawUI");
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
            string data = this.MyChannel.ReadText();
            string[] split = data.Split(new Char[] { '~' });
            if (split.Length != 4)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("UI Error: Can't parse parameters coordinates");
                return;
            }
            Coordinates source;
            int Width, Highth;
            source = new Coordinates(int.Parse(split[0]), int.Parse(split[1]));
            Width = int.Parse(split[2]);
            Highth = int.Parse(split[3]);
            int BufferCellSize = 14; // sizeof(char) + sizeof(int) * 3;
            byte[] RawData = new byte[Width * Highth * BufferCellSize];
            this.MyChannel.SendCommand(SecureChannel.OK);
            int ReadLength = this.MyChannel.ReadData(RawData, 0, RawData.Length);
            if (ReadLength != RawData.Length)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("RawUI Error: Can't parse parameters Rawdata");
                return;
            }
            BufferCell[,] results = new BufferCell[Width, Highth];
            int k = 0;
            for (int i = 0; i < Highth; i++)
            {
                for (int j = 0; j < Width; j++)
                {
                    results[i, j].Character = BitConverter.ToChar(RawData, k);
                    k += 2;
                    results[i, j].ForegroundColor = (ConsoleColor)(BitConverter.ToInt32(RawData, k));
                    k += 4;
                    results[i, j].BackgroundColor = (ConsoleColor)(BitConverter.ToInt32(RawData, k));
                    k += 4;
                    results[i, j].BufferCellType = (BufferCellType)(BitConverter.ToInt32(RawData, k));
                    k += 4;
                }
            }
            try
            {
                this.OutterUI.RawUI.SetBufferContents(source, results);
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void PromptForChoice()
        {
            this.MyChannel.SendCommand(SecureChannel.OK);
            int NumberofChoice = this.MyChannel.ReceiveCommand();
            if (NumberofChoice < 0)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
            string data = this.MyChannel.ReadText();
            string[] split = data.Split(new Char[] { '~' });; 
            if (split.Length != (NumberofChoice*2 + 3))
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("UI Error: Can't parse parameters");
                return;
            }
            string caption, message;
            int defaultchoice = 0, userchoice = 0;
            Collection<ChoiceDescription> choices = new Collection<ChoiceDescription>();
            ChoiceDescription choiceitem;
            caption = split[0];
            message = split[1];
            defaultchoice = int.Parse(split[2]);
            int j = 0;
            for (int i = 0; i < NumberofChoice; i++)
            {
                j = i * 2 + 3;
                choiceitem = new ChoiceDescription(split[j],split[j+1]);
                choices.Add(choiceitem); 
            }
            try
            {
                userchoice = this.OutterUI.PromptForChoice(caption,message,choices,defaultchoice);
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            } 
            if (userchoice <0 || userchoice >= NumberofChoice )
                this.MyChannel.SendCommand(SecureChannel.Error);
            else
                this.MyChannel.SendCommand(userchoice);
        }

        private void WriteProgress()
        {
            this.MyChannel.SendCommand(SecureChannel.OK);
            string data = this.MyChannel.ReadText();
            string[] split = data.Split(new Char[] { '~' }); ;
            if (split.Length != 9)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("UI Error: Can't parse parameters");
                return;
            }
            ProgressRecord record;
            long sourceId;
            sourceId = long.Parse(split[0]);
            record = new ProgressRecord(int.Parse(split[2]), split[1], split[8]);
            record.CurrentOperation = split[3];
            if (string.IsNullOrEmpty(split[4]))
                record.ParentActivityId = 0;
            else
                record.ParentActivityId = int.Parse(split[4]);
            if (string.IsNullOrEmpty(split[5]))
                record.PercentComplete = 0;
            else
                record.PercentComplete = int.Parse(split[5]);
            if (string.IsNullOrEmpty(split[6]))
                record.RecordType =  ProgressRecordType.Processing;
            else
                record.RecordType = (ProgressRecordType)int.Parse(split[6]);
            if (string.IsNullOrEmpty(split[7]))
                record.SecondsRemaining = 60;
            else
                record.SecondsRemaining = int.Parse(split[7]);
            try
            {
                    this.OutterUI.WriteProgress(sourceId, record);
            }
            catch (Exception ex)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            this.MyChannel.SendCommand(SecureChannel.OK);
        }

        private void PromptForCredential()
        {
            this.MyChannel.SendCommand(SecureChannel.OK);
            string data = this.MyChannel.ReadText();
            string[] split = data.Split(new Char[] { '~' }); ;
            if (split.Length != 6)
            {
                this.MyChannel.SendCommand(SecureChannel.Error);
                this.OutterUI.WriteErrorLine("UI Error: Can't parse parameters");
                return;
            }
            string caption, message, username, targetname;
            PSCredentialTypes ctype;
            PSCredentialUIOptions option;
            PSCredential UserCre = null;
            byte[] Buffer = new byte[1];
            caption = split[0];
            message = split[1];
            username = split[2];
            targetname = split[3];
            ctype = (PSCredentialTypes)int.Parse(split[4]);
            option = (PSCredentialUIOptions)int.Parse(split[5]);
            try
            {
                UserCre = this.OutterUI.PromptForCredential(caption, message, username, targetname, ctype, option);
            }
            catch (Exception ex)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                if (this.canThrowExcexption)
                    throw;
                else
                    this.OutterUI.WriteErrorLine(ex.Message);
                return;
            }
            if (UserCre == null)
            {
                this.MyChannel.WriteText(SecureChannel.ErrorLine);
                return; 
            }
            this.MyChannel.WriteText(UserCre.UserName + "~" + SecureStringHelper.GetStringFromSecureString(UserCre.Password));
        }

    }
}
