using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Net.Security;
using System.Security.Principal;
using System.Security.Authentication;

namespace PowerShellRemoting
{
    internal class NegotiateChannel : SecureChannel
    {
        internal NegotiateChannel(NetworkStream stream)
        {
            if (stream == null)
                throw new ArgumentNullException("stream");
            this.InnerStream = stream;
            this.Buffer = new byte[BufferSize];
        }

        private NetworkStream InnerStream = null;

        private NegotiateStream SecureStream = null;

        private bool isAuthenticated = false;

        private const int BufferSize = 5129;

        private byte[] Buffer = null;

        internal override void OpenAsClient(NetworkCredential Cre, string TargetName)
        {
            if (Cre == null)
                throw new ArgumentNullException("Cre");
            try
            {
                this.SecureStream = new NegotiateStream(this.InnerStream, false);
                this.SecureStream.AuthenticateAsClient(Cre,string.Empty ,ProtectionLevel.EncryptAndSign, TokenImpersonationLevel.Impersonation);
                if (this.SecureStream.IsAuthenticated && this.SecureStream.IsEncrypted && this.SecureStream.IsSigned && this.SecureStream.CanRead && this.SecureStream.CanWrite)
                {
                    this.isAuthenticated = true;
                }
                else
                    this.isAuthenticated = false;
            }
            catch (InvalidCredentialException)
            {
                this.isAuthenticated = false;
            }
            catch (AuthenticationException)
            {
                this.isAuthenticated = false;
            }
        }

        internal override void OpenAsServer()
        {
            try
            {
                this.SecureStream = new NegotiateStream(this.InnerStream, false);
                this.SecureStream.AuthenticateAsServer(CredentialCache.DefaultNetworkCredentials, ProtectionLevel.EncryptAndSign, TokenImpersonationLevel.Impersonation);
                if (this.SecureStream.IsAuthenticated && this.SecureStream.IsEncrypted && this.SecureStream.IsSigned && this.SecureStream.CanRead && this.SecureStream.CanWrite)
                {
                    this.isAuthenticated = true;
                }
                else
                    this.isAuthenticated = false;
            }
            catch (InvalidCredentialException)
            {
                this.isAuthenticated = false;
            }
            catch (AuthenticationException)
            {
                this.isAuthenticated = false;
            }
        }

        internal override void Close()
        {
            if (this.SecureStream != null)
                this.SecureStream.Close();
            this.Buffer = null;
        }

        internal override bool IsAuthenticated
        {
            get
            {
                return this.isAuthenticated;
            }
        }

        internal override string Name
        {
            get { return "NegotiateChannel"; }
        }

        internal override WindowsIdentity Identity
        {
            get 
            {
                if (!this.isAuthenticated)
                    return null;
                WindowsIdentity Id = null;
                try
                {
                    Id = (WindowsIdentity)this.SecureStream.RemoteIdentity;
                }
                catch
                {
                    Id = null;
                }
                return Id;
            }
        }

        internal override int ReadData(byte[] buffer, int offset, int count)
        {
            if (this.SecureStream == null || !this.isAuthenticated)
                throw new InvalidOperationException("Read SecureStream");
            return this.SecureStream.Read(buffer, offset, count);
        }

        internal override void WriteData(byte[] buffer, int offset, int count)
        {
            if (this.SecureStream == null || !this.isAuthenticated)
                throw new InvalidOperationException("Write SecureStream");
            this.SecureStream.Write(buffer, offset, count);
        }

        internal override string ReadText()
        {
            StringBuilder sb = new StringBuilder();
            int Length = 0;
            do
            {
                Length = this.ReadData(Buffer, 0, BufferSize);
                sb.Append(Encoding.UTF8.GetString(Buffer, 0, Length));
            }
            while (Length == BufferSize);
            return sb.ToString();

        }

        internal override void WriteText(string message)
        {
            int Start = 0;
            int Length = 0;
            int BlockSize = BufferSize / 2;
            string CurrentSubString = string.Empty;
            byte[] buffer = null;
            do
            {
                if ((Start + BlockSize) > message.Length)
                    Length = message.Length - Start;
                else
                    Length = BlockSize;
                CurrentSubString = message.Substring(Start, Length);
                Start += Length;
                buffer = Encoding.UTF8.GetBytes(CurrentSubString);
                this.WriteData(buffer, 0, buffer.Length);
            }
            while (Length == BlockSize);
        }

        internal override void SendCommand(int cmd)
        {
            if (this.SecureStream == null || !this.isAuthenticated)
                throw new InvalidOperationException("SendCommand");
            byte[] Command = BitConverter.GetBytes(cmd);
            this.WriteData(Command, 0, Command.Length);
        }

        internal override int ReceiveCommand()
        {
            if (this.SecureStream == null || !this.isAuthenticated)
                throw new InvalidOperationException("ReceiveCommand");
            byte[] Command = new byte[sizeof(int)];
            this.ReadData(Command, 0, Command.Length);
            return BitConverter.ToInt32(Command, 0);
        }

        internal override bool SetReadTimeout(int milliseconds)
        {
            if (!this.SecureStream.CanTimeout || (milliseconds <= 0 && milliseconds != System.Threading.Timeout.Infinite))
                return false;
            this.SecureStream.ReadTimeout = milliseconds;
            return true;
        }

        internal override bool UnsetReadTimeout()
        {
            if (!this.SecureStream.CanTimeout)
                return false;
            this.SecureStream.ReadTimeout =  System.Threading.Timeout.Infinite;
            return true;
        }

        internal override bool SetWriteTimeout(int milliseconds)
        {
            if (!this.SecureStream.CanTimeout || (milliseconds <= 0 && milliseconds != System.Threading.Timeout.Infinite))
                return false;
            this.SecureStream.WriteTimeout = milliseconds;
            return true;
        }

        internal override bool UnsetWriteTimeout()
        {
            if (!this.SecureStream.CanTimeout)
                return false;
            this.SecureStream.WriteTimeout = System.Threading.Timeout.Infinite;
            return true;
        }

    }
}
