using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Text;
using System.Management.Automation;
using System.Management.Automation.Runspaces;
using System.Threading;
using System.Diagnostics;


namespace PowerShellRemoting
{
    internal class Executor : Object
    {
        [Flags]
        internal enum ExecutionOptions
        {
            AddOutputter = 1,
            AddToHistory = 2,
            None = 0,
            ReadInputObjects = 4
        }

        private bool cancelled;
        private object instanceStateLock;
        private bool isPromptFunctionExecutor;
        private bool useNestedPipelines;
        private ServerPSHost parent;
        private Pipeline pipeline;

        public const int EventEngineError = 3001;


        internal Executor(ServerPSHost parent, bool useNestedPipelines,bool isPromptFunctionExecutor)
        {
            this.instanceStateLock = new object();
            if (parent == null)
                throw new System.ArgumentNullException("parent");
            this.parent = parent;
            this.useNestedPipelines = useNestedPipelines;
            this.isPromptFunctionExecutor = isPromptFunctionExecutor;
            this.Reset();
        }

        internal Pipeline CurrentPipline
        {
            get 
            { 
                return this.pipeline; 
            }
        }

        private void Reset()
        {
            lock (this.instanceStateLock)
            {
                this.pipeline = null;
                this.cancelled = false;
            }
        }

        internal Pipeline CreatePipeline()
        {
            if (this.useNestedPipelines)
                return this.parent.InnerRunspace.CreateNestedPipeline();
            else
                return this.parent.InnerRunspace.CreatePipeline();
        }

        internal Pipeline CreatePipeline(string command, bool addToHistory)
        {
            if (string.IsNullOrEmpty(command))
                throw new System.ArgumentNullException("command");
            Pipeline pipeline1;
            if (this.useNestedPipelines)
                pipeline1 = this.parent.InnerRunspace.CreateNestedPipeline(command,addToHistory);
            else
                pipeline1 = this.parent.InnerRunspace.CreatePipeline(command, addToHistory);
            if (pipeline1.Commands.Count != 1)
                throw new System.ArgumentException("No command in pipline");
            return pipeline1;
        }


        internal Collection<PSObject> ExecuteCommand(string command)
        {
            Collection<PSObject> collection1 = null;
            Exception exception1 = null;
            collection1 = this.ExecuteCommand(command, out exception1, ExecutionOptions.None);
            if (exception1 != null)
            {
                this.parent.Log.WriteEntry(string.Format("{0}\n\r{1}", exception1.Message, exception1.StackTrace), EventLogEntryType.Error,Executor.EventEngineError);
            }
            else if (collection1 == null|| collection1.Count == 0)
            {
                this.parent.Log.WriteEntry("No objects in result",EventLogEntryType.Error,Executor.EventEngineError);
            }
            return collection1;
        }

        internal Collection<PSObject> ExecuteCommand(string command, out Exception exceptionThrown, ExecutionOptions options)
        {
            if (string.IsNullOrEmpty(command))
                throw new System.ArgumentNullException("command");
            Pipeline pipeline1 = this.CreatePipeline(command,(options & ExecutionOptions.AddToHistory)> ExecutionOptions.None);
            Collection<PSObject> collection1 = this.ExecuteCommandHelper(pipeline1, out exceptionThrown, options);
            return collection1;
        }

        private Collection<PSObject> ExecuteCommandHelper(Pipeline tempPipeline, out Exception exceptionThrown, Executor.ExecutionOptions options)
        {
            if (tempPipeline == null)
                throw new System.ArgumentNullException("tempPipline");
            exceptionThrown = null;
            Collection<PSObject> collection1 = null;
            if ((options & Executor.ExecutionOptions.AddOutputter) > Executor.ExecutionOptions.None)
            {
                if (tempPipeline.Commands.Count == 1)
                    tempPipeline.Commands[0].MergeMyResults(PipelineResultTypes.Error, PipelineResultTypes.Output);
                tempPipeline.Commands.Add("out-default");
            }
            Executor executor1 = this.parent.CurrentExecutor;
            this.parent.CurrentExecutor = this;
            lock (this.instanceStateLock)
            {
                if (this.pipeline != null)
                    throw new System.InvalidOperationException("Another pipline is in use");
                this.pipeline = tempPipeline;
            }
            try
            {
                collection1 = tempPipeline.Invoke();
            }
            catch (Exception exception1)
            {
                ServerPSHost.CheckForSevereException(exception1);
                exceptionThrown = exception1;
            }
            finally
            {
                //this.parent.ui.ResetProgress();
                this.parent.CurrentExecutor = executor1;
                this.Reset();
            }
            return collection1;
        }

        internal string ExecuteCommandAndGetResultAsString(string command, out Exception exceptionThrown)
        {
            exceptionThrown = null;
            string text1 = null;
            Collection<PSObject> collection1 = this.ExecuteCommand(command, out exceptionThrown, ExecutionOptions.None);
            if (exceptionThrown != null)
            {
                this.parent.Log.WriteEntry(string.Format("{0}\n\r{1}", exceptionThrown.Message, exceptionThrown.StackTrace), EventLogEntryType.Error,Executor.EventEngineError);
            }
            else if ((collection1 == null) || (collection1.Count == 0))
            {
                this.parent.Log.WriteEntry("No objects in result", EventLogEntryType.Error,Executor.EventEngineError);
            }
            else
            {
                text1 = collection1[0].ToString();
            }
            return text1;
        }

        internal void Cancel()
        {
            lock (this.instanceStateLock)
            {
                if ((this.pipeline == null) || this.cancelled)
                {
                    return;
                }
                if (this.isPromptFunctionExecutor)
                {
                    Thread.Sleep(100);
                }
                this.pipeline.Stop();
                this.cancelled = true;
            }
        }

    }
}
