using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Net.Security;
using System.Security.Principal;
using System.Security.Authentication;
using System.Security.Cryptography;
using System.Security.Cryptography.X509Certificates;
using System.Runtime.InteropServices;

namespace PowerShellRemoting
{
    internal class SSLChannel : SecureChannel
    {
        internal SSLChannel(NetworkStream stream, X509Certificate X509)
        {
            if (stream == null)
                throw new ArgumentNullException("stream");
            this.InnerStream = stream;
            this.x509 = X509;
            this.Buffer = new byte[BufferSize];
        }

        private NetworkStream InnerStream = null;

        private SslStream SecureStream = null;

        private bool isAuthenticated = false;

        private const int BufferSize = 5129;

        private byte[] Buffer = null;

        private X509Certificate x509 = null;


        public X509Certificate RemoteCertificate
        {
            get 
            {
                if (!this.isAuthenticated || this.SecureStream.RemoteCertificate ==null)
                    return null;
                return this.SecureStream.RemoteCertificate;
            }
        }

        private WindowsIdentity Id = null;

        internal override void OpenAsClient(NetworkCredential Cre, string TargetName)
        {
            if (Cre == null)
                throw new ArgumentNullException("Cre");
            if (string.IsNullOrEmpty(TargetName))
                throw new ArgumentNullException("TargetName");
            try
            {
                this.SecureStream = new SslStream(this.InnerStream, false);
                this.SecureStream.AuthenticateAsClient(TargetName);
                if (!this.SecureStream.IsAuthenticated || !this.SecureStream.IsEncrypted || !this.SecureStream.IsSigned || !this.SecureStream.CanRead || !this.SecureStream.CanWrite)
                {
                    this.isAuthenticated = false;
                    return;
                }
                this.isAuthenticated = true;
                this.SetCredential(Cre);
                if (this.ReceiveCommand() != SecureChannel.OK)
                    this.isAuthenticated = false;
                else
                    this.isAuthenticated = true;
            }
            catch (System.IO.IOException)
            {
                this.isAuthenticated = false;
            }
            catch (AuthenticationException)
            {
                this.isAuthenticated = false;
            }
        }

        internal override void OpenAsServer()
        {
            if (this.x509 == null)
                throw new ArgumentNullException("X509");
            try
            {
                this.SecureStream = new SslStream(this.InnerStream, false);
                this.SecureStream.AuthenticateAsServer(this.x509, false, SslProtocols.Default, true);
                if (!this.SecureStream.IsAuthenticated || !this.SecureStream.IsEncrypted || !this.SecureStream.IsSigned || !this.SecureStream.CanRead || !this.SecureStream.CanWrite)
                {
                    this.isAuthenticated = false;
                    return;
                }
                this.isAuthenticated = true;
                NetworkCredential Cre = this.GetCredential();
                if (Cre == null)
                {
                    this.isAuthenticated = false;
                    return;
                }
                this.LogonUser(Cre);
                if (this.Id == null)
                {
                    this.isAuthenticated = false;
                    this.SendCommand(SecureChannel.Error);
                }
                else
                {
                    this.isAuthenticated = true;
                    this.SendCommand(SecureChannel.OK);
                }
            }
            catch (System.IO.IOException)
            {
                this.isAuthenticated = false;
            }
            catch (AuthenticationException)
            {
                this.isAuthenticated = false;
            }
        }

        internal override void Close()
        {
            if (this.SecureStream != null)
                this.SecureStream.Close();
            this.Buffer = null;
        }

        internal override bool IsAuthenticated
        {
            get
            {
                return this.isAuthenticated;
            }
        }

        internal override string Name
        {
            get { return "SSLChannel"; }
        }

        internal override WindowsIdentity Identity
        {
            get 
            {
                if (!this.isAuthenticated || !this.SecureStream.IsServer)
                    return null;
                return this.Id;
            }
        }

        internal override int ReadData(byte[] buffer, int offset, int count)
        {
            if (this.SecureStream == null || !this.isAuthenticated)
                throw new InvalidOperationException("Read SecureStream");
            return this.SecureStream.Read(buffer, offset, count);
        }

        internal override void WriteData(byte[] buffer, int offset, int count)
        {
            if (this.SecureStream == null || !this.isAuthenticated)
                throw new InvalidOperationException("Write SecureStream");
            this.SecureStream.Write(buffer, offset, count);
        }

        internal override string ReadText()
        {
            StringBuilder sb = new StringBuilder();
            int Length = 0;
            do
            {
                Length = this.ReadData(Buffer, 0, BufferSize);
                sb.Append(Encoding.UTF8.GetString(Buffer, 0, Length));
            }
            while (Length == BufferSize);
            return sb.ToString();

        }

        internal override void WriteText(string message)
        {
            int Start = 0;
            int Length = 0;
            int BlockSize = BufferSize / 2;
            string CurrentSubString = string.Empty;
            byte[] buffer = null;
            do
            {
                if ((Start + BlockSize) > message.Length)
                    Length = message.Length - Start;
                else
                    Length = BlockSize;
                CurrentSubString = message.Substring(Start, Length);
                Start += Length;
                buffer = Encoding.UTF8.GetBytes(CurrentSubString);
                this.WriteData(buffer, 0, buffer.Length);
            }
            while (Length == BlockSize);
        }

        internal override void SendCommand(int cmd)
        {
            if (this.SecureStream == null || !this.isAuthenticated)
                throw new InvalidOperationException("SendCommand");
            byte[] Command = BitConverter.GetBytes(cmd);
            this.WriteData(Command, 0, Command.Length);
        }

        internal override int ReceiveCommand()
        {
            if (this.SecureStream == null || !this.isAuthenticated)
                throw new InvalidOperationException("ReceiveCommand");
            byte[] Command = new byte[sizeof(int)];
            this.ReadData(Command, 0, Command.Length);
            return BitConverter.ToInt32(Command, 0);
        }

        internal override bool SetReadTimeout(int milliseconds)
        {
            if (!this.SecureStream.CanTimeout || (milliseconds <= 0 && milliseconds != System.Threading.Timeout.Infinite))
                return false;
            this.SecureStream.ReadTimeout = milliseconds;
            return true;
        }

        internal override bool UnsetReadTimeout()
        {
            if (!this.SecureStream.CanTimeout)
                return false;
            this.SecureStream.ReadTimeout =  System.Threading.Timeout.Infinite;
            return true;
        }

        internal override bool SetWriteTimeout(int milliseconds)
        {
            if (!this.SecureStream.CanTimeout || (milliseconds <= 0 && milliseconds != System.Threading.Timeout.Infinite))
                return false;
            this.SecureStream.WriteTimeout = milliseconds;
            return true;
        }

        internal override bool UnsetWriteTimeout()
        {
            if (!this.SecureStream.CanTimeout)
                return false;
            this.SecureStream.WriteTimeout = System.Threading.Timeout.Infinite;
            return true;
        }

        private NetworkCredential GetCredential()
        {
            string Credential = this.ReadText();
            if (Credential == SecureChannel.EmptyLine || Credential == SecureChannel.ErrorLine)
                return null;
            int domain, password;
            domain = Credential.IndexOf('\\');
            password = Credential.IndexOf('/');
            if (domain == -1 || password == -1)
                return null;
            // Domain\user/1234
            // 0123456789012345  
            NetworkCredential Cre = new NetworkCredential(
                Credential.Substring(domain + 1, password - domain - 1),
                Credential.Substring(password + 1, Credential.Length - password - 1),
                Credential.Substring(0,domain));
            return Cre;
        }

        private void SetCredential(NetworkCredential Cre)
        {
            string Credential = Cre.Domain + "\\" + Cre.UserName + "/" + Cre.Password;
            this.WriteText(Credential);
        }

        private void LogonUser(NetworkCredential Cre)
        {
            if (Cre == null)
                throw new ArgumentNullException("Cre");
            SafeTokenHandle TokenHandle = new SafeTokenHandle(IntPtr.Zero);
            if (NativeMethods.LogonUser(
                Cre.UserName, Cre.Domain, Cre.Password,
                LogonTypes.Network,
                LogonProviders.Default,
                ref TokenHandle))
            {
                if (TokenHandle.IsInvalid)
                {
                    throw new AuthenticationException("Invalid Token");
                }
                else
                {
                    this.Id = TokenHandle.GetWindowsIdenty();
                }
            }
            else
            {
                throw new InvalidCredentialException("LogonUser Win32Error: " + Marshal.GetLastWin32Error().ToString() + "\nUserName: " + Cre.UserName + "\nDomain: " + Cre.Domain);
            }
        }

    }

}
