#define DebugServer
using System;
using System.Collections;
using System.Collections.ObjectModel;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Management.Automation;
using System.Management.Automation.Host;
using System.Management.Automation.Runspaces;
using System.Globalization;
using System.IO;
using System.Threading;
using System.Diagnostics;
using System.Security.Principal;
using System.Security.Cryptography.X509Certificates;


namespace PowerShellRemoting
{
    internal class ServerPSHost: PSHost, IDisposable
    {
        internal ServerPSHost(TcpClient client, PowerShellRemotingService parent, string message)
        {
            if (parent == null)
                throw new System.ArgumentNullException("parent");
            if (client == null)
                throw new System.ArgumentNullException("client");
            if (string.IsNullOrEmpty(message))
                throw new ArgumentNullException("message");
            this.Client = client;
            this.log = parent.EventLog;
            this.debug = parent.Debug;
            this.hostGlobalLock = new object();
            this.version = new Version(0, 2, 9, 0);
            this.instanceId = Guid.NewGuid();
            this.InputLoopStack = new Stack<InputLoop>();
            this.isDisposed = false;
            this.shouldExit = false;
            this.isRunningAsync = false;
            this.acl = parent.UserACL;
            this.Message = message;
            this.x509 = parent.X509Cer;
        }

        public const int NormalExit = 0;
        public const int ServiceShutDown = 1;
        public const int ClientShutDown = 2;
        public const int ErrorChannelInitialization = 100;
        public const int ErrorChannelSocket = 101;
        public const int ErrorChannelUserBanned = 102;
        public const int ErrorChannelOther = 103;
        public const int ErrorRunspaceInitialization = 200;
        public const int ErrorRunspace = 201;
        public const int ErrorUI = 301;
        public const int ErrorRawUI = 302;
        public const int ErrorOtherGeneral = 401;

        public const int EventExit = 2001;
        public const int EventLoginFailed = 2002;
        public const int EventLoginError = 2003;
        public const int EventLoginSuccess = 2004;
        public const int EventInfo = 2020;
        public const int EventErrorNetwork = 2021;
        public const int EventErrorUnknown = 2022;


        private bool isDisposed;
        private bool debug;
        private string Message = string.Empty;
        internal object hostGlobalLock;
        private string acl;
        private X509Certificate x509;
        private Stack<InputLoop> InputLoopStack;

        private Executor currentExecutor;
        internal Executor CurrentExecutor
        {
            get { return this.currentExecutor; }
            set { this.currentExecutor = value; }
        }

        private WindowsIdentity Id;
        internal WindowsIdentity RemoteId
        {
            get { return this.Id; }
        }

        private bool shouldExit;
        internal bool ShouldExit
        {
            get { return this.shouldExit; }
        }

        private bool isRunningPromptLoop;
        internal bool IsRunningPromptLoop
        {
            get { return this.isRunningPromptLoop; }
            set { this.isRunningPromptLoop = value; }
        }
        private bool isRunningAsync;
        internal bool IsRunningAsync
        {
            get { return this.isRunningAsync; }
            set { this.isRunningAsync = value; }
        }

        private UdpClient ServerCtrlListener = null; 
        private IPEndPoint ClientUdpIPEndPoint = null;

        private TcpClient Client = null;
        public TcpClient ServerTcpClient
        {
            get { return this.Client; }
        }

        private NetworkStream InnerStream = null;
        private SecureChannel myChannel = null;
        public SecureChannel MyChannel
        {
            get { return this.myChannel; }
        }
        
        private EventLog log;
        public EventLog Log
        {
            get { return this.log; }
        }

        public override CultureInfo CurrentCulture
        {
            get { return originalCultureInfo; }
        }
        private CultureInfo originalCultureInfo = System.Threading.Thread.CurrentThread.CurrentCulture;

        public override CultureInfo CurrentUICulture
        {
            get { return originalUICultureInfo; }
        }
        private CultureInfo originalUICultureInfo = System.Threading.Thread.CurrentThread.CurrentUICulture;

        public override void EnterNestedPrompt()
        {
            Executor executor1 = this.currentExecutor;
            try
            {
                bool flag1 = executor1 != null;
                this.RunNewInputLoop(flag1);
            }
            finally
            {
                this.currentExecutor = executor1;
            }
        }

        public override void ExitNestedPrompt()
        {
            lock (this.hostGlobalLock)
            {
                this.ExitCurrentLoop();
            }
        }

        private Guid instanceId;
        public override Guid InstanceId
        {
            get { return instanceId; }
        }

        public override string Name
        {
            get { return "ServerPSHost"; }
        }

        public override void NotifyBeginApplication()
        {
            throw new NotSupportedException("Legacy programs are not supported!");
        }

        public override void NotifyEndApplication()
        {
            return;
        }

        public override void SetShouldExit(int exitCode)
        {
            this.shouldExit = true;
            if (this.debug)
                this.log.WriteEntry("ServerPSHost ExitCode: " + exitCode.ToString(), EventLogEntryType.Information, ServerPSHost.EventExit);
            if (exitCode != 10 && exitCode != 101)
            {
                if (this.myChannel != null && this.myChannel.IsAuthenticated)
                {
                    try
                    {
                        this.myChannel.SendCommand(SecureChannel.Exit);
                    }
                    catch (Exception ex)
                    {
                        if (this.debug)
                        this.log.WriteEntry(string.Format("{0}\n\r{1}", ex.Message, ex.StackTrace), EventLogEntryType.Error, ServerPSHost.EventErrorUnknown);
                    }
                }
            }
        }

        public override PSHostUserInterface UI
        {
            get { return ServerUI; }
        }
        private ServerPSHostUserInterface ServerUI;

        public override Version Version
        {
            get { return this.version; }
        }
        private Version version;

        private Runspace innerRunspace;
        internal Runspace InnerRunspace
        {
            get { return this.innerRunspace; }
        }

        internal static void CheckForSevereException(Exception e)
        {
            if ((e is AccessViolationException) || (e is StackOverflowException))
            {
                Environment.FailFast(e.Message);
            }
        }

        internal void Run(Object obj)
        {
            bool Running = true;
            if (obj != null)
                Running = (bool)obj;
            Thread BreakListener = null;
            try
            {
                if (this.ClientUdpIPEndPoint != null)
                {
                    BreakListener = new Thread(new ThreadStart(this.CtrlListener));
                    BreakListener.Start();
                }
                while (!this.shouldExit && Running)
                {
                    this.InitializeRunspace();
                    if (this.shouldExit)
                        break;
                    WindowsImpersonationContext context = null;
                    try
                    {
                        context = this.Id.Impersonate();
                        Executor initExcutor = new Executor(this, false, false);
                        this.LoadProfile(initExcutor);
                        this.RunInitialScript(initExcutor);
                        initExcutor = null;
                        this.WelcomeClient();
                        this.EnterNestedPrompt();
                    }
                    finally
                    {
                        if (context != null)
                            context.Undo();
                        Id.Dispose();
                        Id = null;
                        context.Dispose();
                        context = null;
                    }
                    this.innerRunspace.Close();
                    this.innerRunspace = null;
                }
            }
            catch (Exception ex)
            {
                this.log.WriteEntry(string.Format("{0}\n\r{1}", ex.Message, ex.StackTrace), EventLogEntryType.Error,ServerPSHost.EventErrorUnknown);
            }
            finally
            {
                if ((BreakListener != null) && (BreakListener.IsAlive))
                {
                    BreakListener.Abort();
                    BreakListener = null;
                }
                this.Dispose();
           }
        }

        internal void InitializeChannel()
        {
            int ChannelType = 0;
            try
            {
                this.InnerStream = Client.GetStream();
                if (this.Message == "~NegotiateChannel")
                {
                    this.myChannel = new NegotiateChannel(this.InnerStream);
                    ChannelType = 1;
                }
                else if (this.Message == "~SSLChannel")
                {
                    this.myChannel = new SSLChannel(this.InnerStream, this.x509);
                    ChannelType = 2;
                }
                else
                {
                    this.log.WriteEntry("Unknown Client SecureChannel " + this.Message, EventLogEntryType.Error, ServerPSHost.EventLoginFailed);
                    this.SetShouldExit(ServerPSHost.ErrorChannelInitialization);
                    return;
                }
                this.myChannel.OpenAsServer();
                if (!this.myChannel.IsAuthenticated)
                {
                    if (this.debug)
                        this.log.WriteEntry("Client (" + ((IPEndPoint)this.Client.Client.RemoteEndPoint).Address.ToString() + ") not authenticated.", EventLogEntryType.Error, ServerPSHost.EventLoginFailed);
                    this.SetShouldExit(ServerPSHost.ErrorChannelInitialization);
                    return;
                }
                this.Id = this.myChannel.Identity;
                if (this.Id == null)
                {
                    if (this.debug)
                        this.log.WriteEntry("Unable to Get Remote Identity of Client (" + ((IPEndPoint)this.Client.Client.RemoteEndPoint).Address.ToString() + ").", EventLogEntryType.Error, ServerPSHost.EventLoginError);
                    this.SetShouldExit(ServerPSHost.ErrorChannelInitialization);
                    return;
                }
                if (ChannelType == 2)
                {
                    this.myChannel.SendCommand(SecureChannel.OK);
                    if (this.myChannel.ReceiveCommand() != SecureChannel.OK)
                    {
                        if (this.debug)
                            this.log.WriteEntry("Remote Client rejected server X509 certificate and disconnected from server.", EventLogEntryType.Information, ServerPSHost.EventLoginError);
                        this.SetShouldExit(ServerPSHost.ErrorChannelInitialization);
                        return;
                    }
                }
                if (this.debug)
                    this.log.WriteEntry("Client (" + ((IPEndPoint)this.Client.Client.RemoteEndPoint).Address.ToString() + ";" + this.myChannel.Identity.Name + ") authenticated.", EventLogEntryType.Information, ServerPSHost.EventLoginSuccess);
                IPEndPoint ServerIPEndPoint = new IPEndPoint(((IPEndPoint)(this.Client.Client.LocalEndPoint)).Address, 0);
                this.ServerCtrlListener = new UdpClient(ServerIPEndPoint);
                if (this.ServerCtrlListener != null)
                    this.myChannel.SendCommand(((IPEndPoint)(this.ServerCtrlListener.Client.LocalEndPoint)).Port);
                else
                    this.myChannel.SendCommand(-1);
                int command = this.myChannel.ReceiveCommand();
                if (command<0 || command < IPEndPoint.MinPort || command > IPEndPoint.MaxPort)
                {
                    this.ClientUdpIPEndPoint = null;
                    this.ServerCtrlListener = null;
                }
                else
                {
                    this.ClientUdpIPEndPoint = new IPEndPoint(((IPEndPoint)(this.Client.Client.RemoteEndPoint)).Address, command);
                    this.ServerCtrlListener.Connect(this.ClientUdpIPEndPoint);
                    if (this.debug)
                        this.log.WriteEntry("Server UDP Port;" + ((IPEndPoint)(this.ServerCtrlListener.Client.LocalEndPoint)).Port.ToString() + "\nClient UDP Port;" + command.ToString(), EventLogEntryType.Information, ServerPSHost.EventInfo);
                    }
                this.ServerUI = new ServerPSHostUserInterface(this);
            }
            catch (Exception ex)
            {
                ServerPSHost.CheckForSevereException(ex);
                this.log.WriteEntry(string.Format("{0}\n\r{1}", ex.Message, ex.StackTrace), EventLogEntryType.Error,ServerPSHost.EventErrorUnknown);
                this.SetShouldExit(ServerPSHost.ErrorChannelInitialization);
            }
        }

        private void WelcomeClient()
        {
            this.ServerUI.WriteLine();
            this.ServerUI.WriteLine(ConsoleColor.White, ConsoleColor.Black, "PowerShell Remoting Server " + this.version.ToString());
            this.ServerUI.WriteLine(ConsoleColor.White, ConsoleColor.Black, "by http://MSHForFun.blogspot.com/ 2006");
            this.ServerUI.WriteLine();
        }

        internal void UserBanned()
        {
            this.ServerUI.WriteLine();
            this.ServerUI.WriteErrorLine("You are NOT allowed to login from this IP. Contact Administrator For details.");
            this.ServerUI.WriteLine();
            this.SetShouldExit(ServerPSHost.ErrorChannelUserBanned);
        }

        private void InitializeRunspace()
        {
            try
            {
                this.innerRunspace = RunspaceFactory.CreateRunspace(this);
                this.innerRunspace.Open();
            }
            catch (Exception ex)
            {
                ServerPSHost.CheckForSevereException(ex);
                this.ReportException(ex);
                this.SetShouldExit(ServerPSHost.ErrorRunspaceInitialization);
            }
        }

        private void LoadProfile(Executor ProfileExcutor)
        {
            if (ProfileExcutor == null)
                throw new ArgumentNullException("ProfileExcutor");
            string MyDoc = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
            string path = PowerShellPath();
            if (!string.IsNullOrEmpty(path))
            {
                string AllUserProfilePath = path + "\\profile.ps1";
                string AllUserShellProfilePath = path+ "\\PowerShellRemoting_profile.ps1";
                this.RunProfile(AllUserProfilePath,ProfileExcutor);
                this.RunProfile(AllUserShellProfilePath, ProfileExcutor);
            }
            string ProfilePath = MyDoc + "\\WindowsPowerShell\\profile.ps1";
            string ShellProfilePath = MyDoc + "\\WindowsPowerShell\\PowerShellRemoting_profile.ps1";
            this.RunProfile(ProfilePath, ProfileExcutor);
            this.RunProfile(ShellProfilePath, ProfileExcutor);
        }

        private string PowerShellPath()
        {
            Microsoft.Win32.RegistryKey software = null, microsoft = null, powershell = null, v1 = null, PowerShellEngine = null;
            string path;
            try
            {
                software = Microsoft.Win32.Registry.LocalMachine.OpenSubKey("Software");
                microsoft = software.OpenSubKey("Microsoft");
                powershell = microsoft.OpenSubKey("PowerShell");
                if (powershell == null)
                {
                    return string.Empty;
                }
                v1 = powershell.OpenSubKey("1");
                PowerShellEngine = v1.OpenSubKey("PowerShellEngine");
                path = (string)PowerShellEngine.GetValue("ApplicationBase", "");
                if (path == "")
                    return string.Empty;
                else
                    return path;
            }
            finally
            {
                if (PowerShellEngine != null)
                    PowerShellEngine.Close();
                if (v1 != null)
                    v1.Close();
                if (powershell != null)
                    powershell.Close();
                if (microsoft != null)
                    microsoft.Close();
                if (software != null)
                    software.Close();
            }

        }


        private void RunInitialScript(Executor InitExcutor)
        {
            if (InitExcutor == null)
                throw new ArgumentNullException("InitExcutor");
            string DocHome = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
            StringBuilder sb = new StringBuilder();
            Exception ex = null;
            if (Directory.Exists(DocHome))
                sb.Append("set-location '" + DocHome + "';");
            else
                sb.Append("set-location 'C:\\Documents and Settings\\All Users\\Documents\\'");
            InitExcutor.ExecuteCommand(sb.ToString(), out ex, Executor.ExecutionOptions.AddOutputter);
            if (ex != null)
            {
                ServerPSHost.CheckForSevereException(ex);
                this.ReportException(ex);
            }
            this.innerRunspace.SessionStateProxy.SetVariable("CurrentUser", this.myChannel.Identity);
            this.innerRunspace.SessionStateProxy.SetVariable("UserACL", this.acl);
        }

        
        private void RunProfile(string profileFileName, Executor ProfileExcutor)
        {
            if (!string.IsNullOrEmpty(profileFileName))
            {
                if (ProfileExcutor == null)
                    throw new ArgumentNullException("ProfileExcutor");
                Exception ex = null;
                if (File.Exists(profileFileName))
                {
                    ProfileExcutor.ExecuteCommand(". '" + ServerPSHost.EscapeTickCharacters(profileFileName) + "'", out ex, Executor.ExecutionOptions.AddOutputter);
                    if (this.debug)
                        this.log.WriteEntry("Run profile: " + profileFileName, EventLogEntryType.Information, ServerPSHost.EventInfo);
                }
                if (ex != null)
                {
                    ServerPSHost.CheckForSevereException(ex);
                    this.ReportException(ex);
                }
            }
        }

        internal static string EscapeTickCharacters(string str)
        {
            StringBuilder builder1 = new StringBuilder(str.Length * 2);
            for (int num1 = 0; num1 < str.Length; num1++)
            {
                char ch1 = str[num1];
                if ((ch1 == '`') || (ch1 == '\''))
                {
                    builder1.Append('`');
                }
                builder1.Append(ch1);
            }
            string text1 = builder1.ToString();
            return  text1;
        }

        internal static string EscapeSeperaterCharacters(string str)
        {
            StringBuilder builder1 = new StringBuilder(str.Length);
            for (int num1 = 0; num1 < str.Length; num1++)
            {
                char ch1 = str[num1];
                if (ch1 == '~')
                    builder1.Append('_');
                else
                    builder1.Append(ch1);
            }
            string text1 = builder1.ToString();
            return text1;
        }

        internal void ReportException(Exception ex)
        {
            if (ex is System.IO.IOException || ex is System.Net.Sockets.SocketException)
            {
                this.log.WriteEntry("Network Error; " + ex.Message, EventLogEntryType.Error,ServerPSHost.EventErrorNetwork);
                this.SetShouldExit(ServerPSHost.ErrorChannelSocket);
                return;
            }

            if (this.myChannel != null && this.myChannel.IsAuthenticated)
            {
                try
                {
                    this.ServerUI.WriteErrorLine(ex.Message);
                }
                catch (Exception ex1)
                {
                    if (!(ex1 is IOException) && !(ex1 is SocketException))
                        this.log.WriteEntry(string.Format("{0}\n\r{1}", ex.Message, ex.StackTrace), EventLogEntryType.Error,ServerPSHost.EventErrorUnknown);
                }
            }
            //this.log.WriteEntry(string.Format("{0}\n\r{1}", ex.Message, ex.StackTrace), EventLogEntryType.Error,ServerPSHost.EventErrorUnknown);
        }

        private void RunNewInputLoop(bool isNested)
        {
            if (this.InputLoopStack.Count == 0x80)
            {
                throw new InvalidOperationException("Too Many Nested Prompts");
            }
            InputLoop loop1 = new InputLoop(this, isNested);
            this.InputLoopStack.Push(loop1);
            loop1.ShouldExit = false;
            loop1.Run();
            InputLoop loop2 = this.InputLoopStack.Pop();
            if (loop1 != loop2)
                throw new InvalidOperationException("top of instance stack does not correspond to the instance pushed");
        }

        private void ExitCurrentLoop()
        {
            if (this.InputLoopStack.Count == 0)
                throw new InvalidOperationException("Current InputLoop Out Of Sync");
            InputLoop loop1 = this.InputLoopStack.Peek();
            loop1.ShouldExit = true;
        }

        private void CtrlListener()
        {
            if (this.ClientUdpIPEndPoint == null)
                return;
            int Command = 0;
            byte[] CommandByte;
            try
            {
                while (!this.shouldExit)
                {
                    CommandByte = ServerCtrlListener.Receive(ref this.ClientUdpIPEndPoint);
                    Command = BitConverter.ToInt32(CommandByte, 0);
                    if (Command == -100)
                    {
                        if (this.currentExecutor != null)
                            this.currentExecutor.Cancel();
                    }
                    else if (Command == -200)
                    {
                        this.SetShouldExit(ServerPSHost.ClientShutDown);
                        break;
                    }
                }
            }
            catch (Exception ex)
            {
                if (ex is System.Threading.ThreadAbortException)
                    return;
                this.ReportException(ex);
            }
            finally
            {
                if (this.ServerCtrlListener != null)
                    this.ServerCtrlListener.Close();
                this.ServerCtrlListener = null;
            }
        }

        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        ~ServerPSHost()
        {
            this.Dispose(false);
        }

        private void Dispose(bool disposing)
        {
            if (!isDisposed)
            {
                if (disposing)
                {
                    if (this.innerRunspace != null)
                    {
                        try
                        {
                            this.innerRunspace.Dispose();
                            this.innerRunspace = null;
                        }
                        catch (InvalidRunspaceStateException exception1)
                        {
                            this.log.WriteEntry(string.Format("{0}\n\r{1}", exception1.Message, exception1.StackTrace), EventLogEntryType.Error, ServerPSHost.EventErrorUnknown);
                        }
                    }
                    if (this.myChannel != null)
                    {
                        this.myChannel.Close();
                        this.myChannel = null;
                    }
                    if (this.InnerStream != null)
                    {
                        this.InnerStream.Close();
                        this.InnerStream = null;
                    }
                    if (this.Client != null)
                    {
                        this.Client.Close();
                        this.Client = null;
                    }
                    if (this.ServerCtrlListener != null)
                    {
                        this.ServerCtrlListener.Close();
                        this.ServerCtrlListener = null;
                    }
                    this.ServerUI = null;
                    this.hostGlobalLock = null;
                }
            }
            this.isDisposed = true;
            this.shouldExit = true;
        }

    }
}
