using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Globalization;
using System.Text;
using System.Text.RegularExpressions;
using System.Security;
using System.Management.Automation;
using System.Management.Automation.Host;
using System.Management.Automation.Runspaces;


namespace PowerShellRemoting
{
    internal class ServerPSHostUserInterface : PSHostUserInterface
    {
        internal ServerPSHostUserInterface(ServerPSHost Parent)
        {
            if (Parent == null)
                throw new ArgumentNullException("Parent");
            this.parent = Parent;
            this.instanceLock = new object();
            this.myRawUi = new ServerPSHostRawUserInterface(this);
        }

        private ServerPSHost parent;
        internal ServerPSHost Parent
        {
            get { return this.parent; }
        }

        private Object instanceLock;

        private enum PromptCommonInputErrors
        {
            None,
            Format,
            Overflow
        }

        public override Dictionary<string, PSObject> Prompt(string caption, string message, Collection<FieldDescription> descriptions)
        {
            if (descriptions == null || descriptions.Count < 1)
                throw new System.ArgumentNullException("description");
            Dictionary<string, PSObject> results = new Dictionary<string, PSObject>();
            lock (this.instanceLock)
            {
                if (!string.IsNullOrEmpty(caption))
                    this.WriteLine(ConsoleColor.White, ConsoleColor.Black, caption);
                if (!string.IsNullOrEmpty(message))
                    this.WriteLine(ConsoleColor.White, ConsoleColor.Black, message);
                int num1 = -1;
                bool cancelled = false;
                foreach (FieldDescription description1 in descriptions)
                {
                    num1++;
                    if (description1 == null)
                        throw new ArgumentNullException("FieldDescription");
                    PSObject obj1 = null;
                    string descriptionName = description1.Name;
                    bool fieldEchoOnPrompt = true;
                    if (string.IsNullOrEmpty(description1.ParameterAssemblyFullName))
                    {
                        throw new ArgumentException(string.Format(CultureInfo.InvariantCulture, "descriptions[{0}].AssemblyFullName", new object[] { num1 }));
                    }
                    Type type1 = ServerPSHostUserInterface.GetTypeFromString(description1.ParameterAssemblyFullName);
                    if (type1 == null)
                    {
                        if (ServerPSHostUserInterface.isSecureType(description1.ParameterTypeName))
                            throw new ArgumentException(description1.ParameterTypeName);
                        type1 = typeof(string);
                    }

                    if (type1.GetInterface(typeof(IList).FullName) != null)
                    {
                        ArrayList list1 = new ArrayList();
                        Type type2 = typeof(object);
                        if (type1.IsArray)
                        {
                            type2 = type1.GetElementType();
                            int num2 = type1.GetArrayRank();
                            if (num2 <= 0)
                            {
                                throw new InvalidOperationException(string.Format(CultureInfo.InvariantCulture, "descriptions[{0}].AssemblyFullName", new object[] { num1 }));
                            }
                        }
                        StringBuilder builder1 = new StringBuilder(descriptionName);
                        builder1.Append("[");
                        while (true)
                        {
                            builder1.Append(string.Format(CultureInfo.InvariantCulture, "{0}]: ", new object[] { list1.Count }));
                            bool endListInput = false;
                            string text5 = this.PromptReadInput(builder1.ToString(), description1, fieldEchoOnPrompt, true, out endListInput, out cancelled);
                            if (cancelled || endListInput)
                            {
                                break;
                            }
                            object obj2 = null;
                            if (this.PromptTryConvertTo(type2, text5, out obj2) == PromptCommonInputErrors.None)
                            {
                                list1.Add(obj2);
                            }
                            builder1.Length = descriptionName.Length + 1;
                        }
                        if (!cancelled)
                        {
                            object obj3 = null;
                            if (LanguagePrimitives.TryConvertTo(list1, type1, out obj3))
                            {
                                obj1 = PSObject.AsPSObject(obj3);
                            }
                            else
                            {
                                obj1 = PSObject.AsPSObject(list1);
                            }
                        }
                    }
                    else
                    {
                        object obj4 = null;
                        if (type1.Equals(typeof(SecureString)))
                        {
                            this.Write(descriptionName);
                            obj4 = this.ReadLineAsSecureString();
                            cancelled = (obj4 == null);
                        }
                        else if (type1.Equals(typeof(PSCredential)))
                        {
                            this.WriteLine(descriptionName);
                            obj4 = this.PromptForCredential(null, null, null, string.Empty);
                            cancelled = (obj4 == null);
                        }
                        else
                        {
                            bool endListInput = false;
                            string text7 = null;
                            do
                            {
                                text7 = this.PromptReadInput(descriptionName, description1, fieldEchoOnPrompt, false, out endListInput, out cancelled);
                            }
                            while (!cancelled && (this.PromptTryConvertTo(type1, text7, out obj4) != PromptCommonInputErrors.None));
                        }
                        if (!cancelled)
                        {
                            obj1 = PSObject.AsPSObject(obj4);
                        }
                    }
                    if (cancelled)
                    {
                        this.WriteLine();
                        results.Clear();
                        break;
                    }
                    results.Add(description1.Name, PSObject.AsPSObject(obj1));
                }
            return results;
            }
        }

        public override int PromptForChoice(string caption, string message, Collection<ChoiceDescription> choices, int defaultChoice)
        {
            if (choices == null)
            {
                throw new ArgumentNullException("choices");
            }
            if (choices.Count == 0)
            {
                throw new ArgumentException("choices");
            }
            if ((defaultChoice < 0) || (defaultChoice >= choices.Count))
            {
                throw new ArgumentOutOfRangeException("defaultChoice");
            }
            int command = 0;
            int NumberofChoice = choices.Count;
            StringBuilder sb = new StringBuilder();
            if (string.IsNullOrEmpty(caption))
                sb.Append("~");
            else
                sb.Append(ServerPSHost.EscapeSeperaterCharacters(caption) + "~");
            if (string.IsNullOrEmpty(message))
                sb.Append("~");
            else
                sb.Append(ServerPSHost.EscapeSeperaterCharacters(message) + "~");
            sb.Append(defaultChoice.ToString() + "~");
            foreach (ChoiceDescription choiceitem in choices)
            {
                sb.Append(ServerPSHost.EscapeSeperaterCharacters(choiceitem.Label) + "~");
                sb.Append(ServerPSHost.EscapeSeperaterCharacters(choiceitem.HelpMessage) + "~");
            }
            sb.Remove(sb.Length - 1, 1);
            lock (this.instanceLock)
            {
                this.parent.MyChannel.SendCommand(SecureChannel.PromptForChoice);
                command = this.parent.MyChannel.ReceiveCommand();
                if (command != SecureChannel.OK)
                    throw new InvalidOperationException("PromptForChoice");
                this.parent.MyChannel.SendCommand(NumberofChoice);
                command = this.parent.MyChannel.ReceiveCommand();
                if (command != SecureChannel.OK)
                    throw new InvalidOperationException("PromptForChoice");
                this.parent.MyChannel.WriteText(sb.ToString());
                this.parent.MyChannel.UnsetReadTimeout();
                command = this.parent.MyChannel.ReceiveCommand();
                //this.parent.MyChannel.SetReadTimeout(5000);
                if (command == SecureChannel.Error)
                    throw new InvalidOperationException("PromptForChoice");
                if ((command < 0) || (command >= NumberofChoice))
                    return defaultChoice;
                else
                    return command;
            } 
        }

        public override PSCredential PromptForCredential(string caption, string message, string userName, string targetName)
        {
            return this.PromptForCredential(caption, message, userName, targetName, PSCredentialTypes.Default, PSCredentialUIOptions.ValidateUserNameSyntax);
        }

        public override PSCredential PromptForCredential(string caption, string message, string userName, string targetName, PSCredentialTypes allowedCredentialTypes, PSCredentialUIOptions options)
        {
            int command = 0;
            string UserName, pass;
            SecureString Password = new SecureString();
            StringBuilder sb = new StringBuilder();
            if (string.IsNullOrEmpty(caption))
                sb.Append("~");
            else
                sb.Append(ServerPSHost.EscapeSeperaterCharacters(caption) + "~");
            if (string.IsNullOrEmpty(message))
                sb.Append("~");
            else
                sb.Append(ServerPSHost.EscapeSeperaterCharacters(message) + "~");
            if (string.IsNullOrEmpty(userName))
                sb.Append("~");
            else
                sb.Append(userName + "~");
            if (string.IsNullOrEmpty(targetName))
                sb.Append("~");
            else
                sb.Append(targetName + "~");
            sb.Append(((int)allowedCredentialTypes).ToString() + "~");
            sb.Append(((int)options).ToString());
            lock (this.instanceLock)
            {
                this.parent.MyChannel.SendCommand(SecureChannel.PromptForCredential);
                command = this.parent.MyChannel.ReceiveCommand();
                if (command != SecureChannel.OK)
                    throw new InvalidOperationException("PromptForCredential");
                this.parent.MyChannel.WriteText(sb.ToString());
                this.parent.MyChannel.UnsetReadTimeout();
                string data = this.parent.MyChannel.ReadText();
                //this.parent.MyChannel.SetReadTimeout(5000);
                if (data == SecureChannel.ErrorLine)
                    return null;
                string[] split = data.Split(new Char[] { '~' });
                if (split.Length != 2)
                {
                    return null;
                }
                UserName = split[0];
                if (string.IsNullOrEmpty(UserName))
                    return null;
                pass = split[1];
                if (string.IsNullOrEmpty(pass))
                    Password = new SecureString();
                else
                {
                    byte[] tmp = Encoding.Unicode.GetBytes(pass);
                    Password = SecureStringHelper.NewSecureString(tmp, tmp.Length);
                }
                pass = string.Empty;
                return new PSCredential(UserName,Password);
            }
        }

        private ServerPSHostRawUserInterface myRawUi;
        public override PSHostRawUserInterface RawUI
        {
            get { return myRawUi; }
        }

        public override string ReadLine()
        {
            string line = string.Empty;
            lock (this.instanceLock)
            {
                this.parent.MyChannel.SendCommand(SecureChannel.ReadLine);
                this.parent.MyChannel.UnsetReadTimeout();
                line = this.parent.MyChannel.ReadText();
                //this.parent.MyChannel.SetReadTimeout(5000);
            }
            if (line == SecureChannel.ErrorLine)
                throw new InvalidOperationException("ReadLine");
            else if (line == SecureChannel.EmptyLine)
                return string.Empty;
            else
                return line;
        }

        public override SecureString ReadLineAsSecureString()
        {
            byte[] TmpBuffer ;
            string Password;
            lock (this.instanceLock)
            {
                this.parent.MyChannel.SendCommand(SecureChannel.ReadLineAsSecureString);
                this.parent.MyChannel.UnsetReadTimeout();
                Password = this.parent.MyChannel.ReadText();
                //this.parent.MyChannel.SetReadTimeout(5000);
                if (Password == SecureChannel.ErrorLine)
                    throw new InvalidOperationException("Invalid SecureString Length");
                if (Password == SecureChannel.EmptyLine)
                    return new SecureString();
            }
            TmpBuffer = Encoding.Unicode.GetBytes(Password);
            SecureString text1 = SecureStringHelper.NewSecureString(TmpBuffer,TmpBuffer.Length);
            if (text1 == null)
                throw new InvalidOperationException("Null SecureString");
            return text1;
        }

        public override void Write(string value)
        {
            if (string.IsNullOrEmpty(value))
                return;
            this.parent.MyChannel.SendCommand(SecureChannel.Write);
            int command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("Unexpected response from client while calling write()");
            this.parent.MyChannel.WriteText(value);
            command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("Unexpected response from client while calling write()");
        }

        public override void Write(ConsoleColor foregroundColor, ConsoleColor backgroundColor, string value)
        {
            if (string.IsNullOrEmpty(value))
                return;
            value = ServerPSHost.EscapeSeperaterCharacters(value);
            string data= ((int)foregroundColor).ToString()+"~"+((int)backgroundColor).ToString()+"~"+value;
            this.parent.MyChannel.SendCommand(SecureChannel.WriteColor);
            int command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("Unexpected response from client while calling writecolor()");
            this.parent.MyChannel.WriteText(data);
            command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("Unexpected response from client while calling writecolor()");
        }

        public override void WriteLine()
        {
            this.parent.MyChannel.SendCommand(SecureChannel.WriteEmptyLine);
            int command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("Unexpected response from client while calling WriteEmptyLine()");
        }

        public override void WriteLine(string value)
        {
            if (string.IsNullOrEmpty(value))
                return;
            this.parent.MyChannel.SendCommand(SecureChannel.WriteLine);
            int command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("Unexpected response from client while calling writeLine()");
            this.parent.MyChannel.WriteText(value);
            command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("Unexpected response from client while calling writeLine()");
        }

        public override void WriteLine(ConsoleColor foregroundColor, ConsoleColor backgroundColor, string value)
        {
            if (string.IsNullOrEmpty(value))
                return;
            value = ServerPSHost.EscapeSeperaterCharacters(value);
            string data = ((int)foregroundColor).ToString() + "~" + ((int)backgroundColor).ToString() + "~" + value;
            this.parent.MyChannel.SendCommand(SecureChannel.WriteLineColor);
            int command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("Unexpected response from client while calling writeLinecolor()");
            this.parent.MyChannel.WriteText(data);
            command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("Unexpected response from client while calling writeLinecolor()");
        }

        public override void WriteDebugLine(string message)
        {
            if (string.IsNullOrEmpty(message))
                return;
            this.parent.MyChannel.SendCommand(SecureChannel.WriteDebugLine);
            int command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("Unexpected response from client while calling writeDebugline()");
            this.parent.MyChannel.WriteText(message);
            command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("Unexpected response from client while calling writeDebugline()");
        }

        public override void WriteErrorLine(string value)
        {
            if (string.IsNullOrEmpty(value))
                return;
            this.parent.MyChannel.SendCommand(SecureChannel.WriteErrorLine);
            int command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("Unexpected response from client while calling writeErrorline()");
            this.parent.MyChannel.WriteText(value);
            command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("Unexpected response from client while calling writeErrorline()");
        }

        public override void WriteVerboseLine(string message)
        {
            if (string.IsNullOrEmpty(message))
                return;
            this.parent.MyChannel.SendCommand(SecureChannel.WriteVerboseLine);
            int command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("Unexpected response from client while calling writeverboseline()");
            this.parent.MyChannel.WriteText(message);
            command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("Unexpected response from client while calling writeverboseline()");
        }

        public override void WriteWarningLine(string message)
        {
            if (string.IsNullOrEmpty(message))
                return;
            this.parent.MyChannel.SendCommand(SecureChannel.WriteWarningLine);
            int command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("Unexpected response from client while calling writeWarningline()");
            this.parent.MyChannel.WriteText(message);
            command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("Unexpected response from client while calling writeWarningline()");
        }

        public override void WriteProgress(long sourceId, ProgressRecord record)
        {
            if (record == null)
                throw new ArgumentNullException("record");
            StringBuilder sb = new StringBuilder();
            int command;
            sb.Append(sourceId.ToString() + "~");
            sb.Append(ServerPSHost.EscapeSeperaterCharacters((record.Activity == null ? string.Empty:record.Activity)) + "~");
            sb.Append(record.ActivityId.ToString() + "~");
            sb.Append(ServerPSHost.EscapeSeperaterCharacters((record.CurrentOperation == null ? string.Empty:record.CurrentOperation)) + "~");
            sb.Append(record.ParentActivityId.ToString() + "~");
            sb.Append(record.PercentComplete.ToString() + "~");
            sb.Append(((int)(record.RecordType)).ToString() + "~");
            sb.Append(record.SecondsRemaining.ToString() + "~");
            sb.Append(ServerPSHost.EscapeSeperaterCharacters((record.StatusDescription == null ? string.Empty:record.StatusDescription)));
            this.parent.MyChannel.SendCommand(SecureChannel.WriteProgress);
            command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("WriteProgress");
            this.parent.MyChannel.WriteText(sb.ToString());
            command = this.parent.MyChannel.ReceiveCommand();
            if (command != SecureChannel.OK)
                throw new InvalidOperationException("WriteProgress");
        }

        private static Type GetTypeFromString(string assemblyQualifiedTypeName)
        {
            Type type1 = Type.GetType(assemblyQualifiedTypeName);
            if (type1 == null)
            {
                int num1 = assemblyQualifiedTypeName.IndexOf(",");
                string text1 = assemblyQualifiedTypeName.Substring(num1 + 2, (assemblyQualifiedTypeName.Length - num1) - 2);
                string text2 = assemblyQualifiedTypeName.Substring(0, num1);
                foreach (System.Reflection.Assembly assembly1 in AppDomain.CurrentDomain.GetAssemblies())
                {
                    if (assembly1.FullName == text1)
                    {
                        return assembly1.GetType(text2);
                    }
                }
            }
            return type1;
        }

        private static bool isSecureType(string typeName)
        {
            if (string.Compare("SecureString", typeName, true, CultureInfo.InvariantCulture) != 0)
            {
                return (string.Compare("MshCredential", typeName, true, CultureInfo.InvariantCulture) == 0);
            }
            return true;
        }

        private string PromptReadInput(string fieldPrompt, FieldDescription desc, bool fieldEchoOnPrompt, bool listInput, out bool endListInput, out bool cancelled)
        {
            if(fieldPrompt == null)
                throw new ArgumentNullException("fieldPrompt");
            if (desc == null)
                throw new ArgumentNullException("desc");
            string text1 = null;
            endListInput = false;
            cancelled = false;
            bool flag1 = false;
            while (!flag1)
            {
                this.Write(ConsoleColor.White,ConsoleColor.Black,fieldPrompt);
                string text2 = null;
                if (fieldEchoOnPrompt)
                {
                    text2 = this.ReadLine();
                }
                else
                {
                    string text3 = SecureStringHelper.GetStringFromSecureString(this.ReadLineAsSecureString());
                    if (text3 == null)
                        throw new InvalidOperationException("ReadLineSafe");
                    text2 = text3;
                }
                if (text2 == null)
                {
                    cancelled = true;
                    break;
                }
                if (text2.StartsWith("!", false, CultureInfo.CurrentCulture))
                {
                    text1 = this.PromptCommandMode(text2, desc, out flag1);
                }
                else
                {
                    if ((text2.Length == 0) && listInput)
                    {
                        endListInput = true;
                    }
                    text1 = text2;
                    break;
                }
            }
            return text1;
        }

        private string PromptCommandMode(string input, FieldDescription desc, out bool inputDone)
        {
            if ((input == null) || !input.StartsWith("!", StringComparison.InvariantCultureIgnoreCase))
                throw new ArgumentException( string.Format(CultureInfo.InvariantCulture, "input should start with {0}", new object[] { "!" }));
            if(desc != null)
                throw new ArgumentNullException("desc");
            string text1 = input.Substring(1);
            inputDone = true;
            if (text1.StartsWith("!", StringComparison.InvariantCultureIgnoreCase))
            {
                return text1;
            }
            if (text1.Length == 1)
            {
                if (text1[0] == '?')
                {
                    if (string.IsNullOrEmpty(desc.HelpMessage))
                    {
                        this.WriteLine("No Help Available for Prompt");
                    }
                    else
                    {
                        this.WriteLine(desc.HelpMessage);
                    }
                }
                else
                {
                    this.WriteLine("Unrecognized Prompt Command: " + input);
                }
                inputDone = false;
                return null;
            }
            if ((text1.Length == 2) && (string.Compare(text1, "\"\"", StringComparison.InvariantCultureIgnoreCase) == 0))
            {
                return string.Empty;
            }
            if (string.Compare(text1, "$null", StringComparison.InvariantCultureIgnoreCase) != 0)
            {
                this.WriteLine("Unrecognized Prompt Command: " + input);
                inputDone = false;
            }
            return null;
        }

        private PromptCommonInputErrors PromptTryConvertTo(Type fieldType, string inputString, out object convertedObj)
        {
           if (fieldType == null)
               throw new ArgumentNullException("fieldType");
            convertedObj = inputString;
            try
            {
                convertedObj = LanguagePrimitives.ConvertTo(inputString, fieldType, CultureInfo.InvariantCulture);
            }
            catch (PSInvalidCastException exception1)
            {
                Exception exception2 = exception1.InnerException;
                this.parent.Log.WriteEntry(string.Format("{0}\n\r{1}", exception1.InnerException.Message, exception1.InnerException.StackTrace), System.Diagnostics.EventLogEntryType.Error);
                if (exception2 != null)
                {
                    if (exception2 is OverflowException)
                    {
                        this.WriteErrorLine(string.Format(CultureInfo.CurrentCulture, "Prompt ({0}:{1}) Parse Overflow", new object[] { fieldType, inputString }));
                        return PromptCommonInputErrors.Overflow;
                    }
                    if (exception2 is FormatException)
                    {
                        if (inputString.Length > 0)
                        {
                            this.WriteErrorLine(string.Format(CultureInfo.CurrentCulture, "Prompt ({0}:{1}) PFormat Error", new object[] { fieldType, inputString }));
                        }
                        return PromptCommonInputErrors.Format;
                    }
                }
            }
            return PromptCommonInputErrors.None;
        }

    }
}
